"""Settings."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING
from urllib.parse import urlparse
import json
import logging
import re

from . import utils

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from pathlib import Path

__all__ = ('TYPE_CHECKSUM', 'TYPE_COMMIT', 'TYPE_DAVINCI', 'TYPE_DIRECTORY', 'TYPE_METADATA',
           'TYPE_NONE', 'TYPE_REGEX', 'TYPE_REPOLOGY', 'LivecheckSettings', 'gather_settings')

log = logging.getLogger(__name__)
TYPE_CHECKSUM = 'checksum'
TYPE_COMMIT = 'commit'
TYPE_DAVINCI = 'davinci'
TYPE_DIRECTORY = 'directory'
TYPE_METADATA = 'metadata'
TYPE_NONE = 'none'
TYPE_REGEX = 'regex'
TYPE_REPOLOGY = 'repology'

SETTINGS_TYPES = {
    TYPE_CHECKSUM, TYPE_COMMIT, TYPE_DAVINCI, TYPE_DIRECTORY, TYPE_METADATA, TYPE_NONE, TYPE_REGEX,
    TYPE_REPOLOGY
}


@dataclass
class LivecheckSettings:
    """All settings."""
    branches: dict[str, str] = field(default_factory=dict)
    custom_livechecks: dict[str, tuple[str, str]] = field(default_factory=dict)
    dotnet_projects: dict[str, str] = field(default_factory=dict)
    """Dictionary of catpkg to project or solution file (base name only)."""
    go_sum_uri: dict[str, str] = field(default_factory=dict)
    """
    Dictionary of catpkg to full URI to ``go.sum`` with ``@PV@`` used for
    where version gets placed.
    """
    type_packages: dict[str, str] = field(default_factory=dict)
    no_auto_update: set[str] = field(default_factory=set)
    """Disable auto-detection of semantic versioning."""
    sha_sources: dict[str, str] = field(default_factory=dict)
    transformations: Mapping[str, Callable[[str], str]] = field(default_factory=dict)
    yarn_base_packages: dict[str, str] = field(default_factory=dict)
    yarn_packages: dict[str, set[str]] = field(default_factory=dict)
    jetbrains_packages: dict[str, bool] = field(default_factory=dict)
    keep_old: dict[str, bool] = field(default_factory=dict)
    gomodule_packages: dict[str, bool] = field(default_factory=dict)
    gomodule_path: dict[str, str] = field(default_factory=dict)
    nodejs_packages: dict[str, bool] = field(default_factory=dict)
    nodejs_path: dict[str, str] = field(default_factory=dict)
    development: dict[str, bool] = field(default_factory=dict)
    composer_packages: dict[str, bool] = field(default_factory=dict)
    composer_path: dict[str, str] = field(default_factory=dict)
    regex_version: dict[str, tuple[str, str]] = field(default_factory=dict)
    restrict_version: dict[str, str] = field(default_factory=dict)
    sync_version: dict[str, str] = field(default_factory=dict)
    stable_version: dict[str, str] = field(default_factory=dict)
    # Settings from command line flag.
    auto_update_flag: bool = False
    debug_flag: bool = False
    development_flag: bool = False
    git_flag: bool = False
    keep_old_flag: bool = False
    progress_flag: bool = False
    # Internal settings.
    restrict_version_process: str = ''

    def is_devel(self, catpkg: str) -> bool:
        """Check if the package is a development version."""
        return self.development.get(catpkg, self.development_flag)


class UnknownTransformationFunction(NameError):
    def __init__(self, tfs: str) -> None:
        super().__init__(f'Unknown transformation function: {tfs}')


def gather_settings(search_dir: Path) -> LivecheckSettings:
    """
    Gather settings from ``livecheck.json`` files in the given directory.

    Raises
    ------
    UnknownTransformationFunction
        If a transformation function is invalid.
    """
    # Prevent circular import.
    import livecheck.special.handlers as sc  # noqa: PLC0415

    branches: dict[str, str] = {}
    custom_livechecks: dict[str, tuple[str, str]] = {}
    dotnet_projects: dict[str, str] = {}
    golang_packages: dict[str, str] = {}
    type_packages: dict[str, str] = {}
    no_auto_update: set[str] = set()
    sha_sources: dict[str, str] = {}
    transformations: dict[str, Callable[[str], str]] = {}
    yarn_base_packages: dict[str, str] = {}
    yarn_packages: dict[str, set[str]] = {}
    jetbrains_packages: dict[str, bool] = {}
    keep_old: dict[str, bool] = {}
    gomodule_packages: dict[str, bool] = {}
    gomodule_path: dict[str, str] = {}
    nodejs_packages: dict[str, bool] = {}
    nodejs_path: dict[str, str] = {}
    development: dict[str, bool] = {}
    composer_packages: dict[str, bool] = {}
    composer_path: dict[str, str] = {}
    regex_version: dict[str, tuple[str, str]] = {}
    restrict_version: dict[str, str] = {}
    sync_version: dict[str, str] = {}
    stable_version: dict[str, str] = {}

    for path in search_dir.glob('**/livecheck.json'):
        log.debug('Opening %s.', path)
        with path.open() as f:
            dn = path.parent
            catpkg = f'{dn.parent.name}/{dn.name}'
            try:
                settings_parsed = json.load(f)
            except json.JSONDecodeError:
                log.exception('Error parsing file %s.', path)
                continue
            if settings_parsed.get('type') is not None:
                type_ = settings_parsed.get('type').lower()
                if type_ == TYPE_REGEX:
                    if settings_parsed.get('url') is None:
                        log.error('No "url" in %s.', path)
                        continue
                    if settings_parsed.get('regex') is None:
                        log.error('No "regex" in %s.', path)
                        continue
                    custom_livechecks[catpkg] = (settings_parsed['url'], settings_parsed['regex'])
                if type_ == TYPE_REPOLOGY:
                    if settings_parsed.get('package') is None:
                        log.error('No "package" in %s.', path)
                        continue
                    custom_livechecks[catpkg] = (settings_parsed.get('package'), '')
                if type_ == TYPE_DIRECTORY:
                    if settings_parsed.get('url') is None:
                        log.error('No "url" in %s.', path)
                        continue
                    custom_livechecks[catpkg] = (settings_parsed.get('url'), '')
                if type_ not in SETTINGS_TYPES:
                    log.error('Unknown "type" in %s.', path)
                else:
                    type_packages[catpkg] = type_

            if settings_parsed.get('branch'):
                check_instance(settings_parsed['branch'], 'branch', 'string', path)
                branches[catpkg] = settings_parsed['branch']
            if 'no_auto_update' in settings_parsed:
                check_instance(settings_parsed['no_auto_update'], 'no_auto_update', 'bool', path,
                               True)  # noqa: FBT003
                no_auto_update.add(catpkg)
            if settings_parsed.get('transformation_function', None):
                tfs = settings_parsed['transformation_function']
                check_instance(settings_parsed['transformation_function'],
                               'transformation_function', 'string', path)
                try:
                    tf: Callable[[str], str] = getattr(sc, tfs)
                except AttributeError:
                    try:
                        tf = getattr(utils, tfs)
                    except AttributeError as e:
                        raise UnknownTransformationFunction(tfs) from e
                transformations[catpkg] = tf
            if settings_parsed.get('sha_source'):
                check_instance(settings_parsed['sha_source'], 'sha_source', 'url', path)
                sha_sources[catpkg] = settings_parsed['sha_source']
            if settings_parsed.get('yarn_base_package'):
                check_instance(settings_parsed['yarn_base_package'], 'yarn_base_package', 'string',
                               path)
                yarn_base_packages[catpkg] = settings_parsed['yarn_base_package']
                if settings_parsed.get('yarn_packages'):
                    check_instance(settings_parsed['yarn_packages'], 'yarn_packages', 'list', path)
                    yarn_packages[catpkg] = set(settings_parsed['yarn_packages'])
            if settings_parsed.get('go_sum_uri'):
                check_instance(settings_parsed['go_sum_uri'], 'go_sum_uri', 'url', path)
                golang_packages[catpkg] = settings_parsed['go_sum_uri']
            if settings_parsed.get('dotnet_project'):
                check_instance(settings_parsed['dotnet_project'], 'dotnet_project', 'string', path)
                dotnet_projects[catpkg] = settings_parsed['dotnet_project']
            if 'jetbrains' in settings_parsed:
                check_instance(settings_parsed['jetbrains'], 'jetbrains', 'bool', path)
                jetbrains_packages[catpkg] = settings_parsed['jetbrains']
            if 'keep_old' in settings_parsed:
                check_instance(settings_parsed['keep_old'], 'keep_old', 'bool', path)
                keep_old[catpkg] = settings_parsed['keep_old']
            if 'gomodule' in settings_parsed:
                check_instance(settings_parsed['gomodule'], 'gomodule', 'bool', path)
                gomodule_packages[catpkg] = settings_parsed['gomodule']
                gomodule_path[catpkg] = ''
                if settings_parsed.get('gomodule_path'):
                    check_instance(settings_parsed['gomodule_path'], 'gomodule_path', 'string',
                                   path)
                    gomodule_path[catpkg] = settings_parsed['gomodule_path']
            if 'nodejs' in settings_parsed:
                check_instance(settings_parsed['nodejs'], 'nodejs', 'bool', path)
                nodejs_packages[catpkg] = settings_parsed['nodejs']
                nodejs_path[catpkg] = ''
                if settings_parsed.get('nodejs_path'):
                    check_instance(settings_parsed['nodejs_path'], 'nodejs_path', 'string', path)
                    nodejs_path[catpkg] = settings_parsed['nodejs_path']
            if 'development' in settings_parsed:
                check_instance(settings_parsed['development'], 'development', 'bool', path)
                development[catpkg] = settings_parsed['development']
            if 'composer' in settings_parsed:
                check_instance(settings_parsed['composer'], 'composer', 'bool', path)
                composer_packages[catpkg] = settings_parsed['composer']
                composer_path[catpkg] = ''
                if settings_parsed.get('composer_path'):
                    check_instance(settings_parsed['composer_path'], 'composer_path', 'string',
                                   path)
                    composer_path[catpkg] = settings_parsed['composer_path']
            if 'pattern_version' in settings_parsed or 'replace_version' in settings_parsed:
                if 'pattern_version' not in settings_parsed:
                    log.error('No "pattern_version" in %s.', path)
                    continue
                if 'replace_version' not in settings_parsed:
                    log.error('No "replace_version" in %s.', path)
                    continue
                check_instance(settings_parsed['pattern_version'], 'pattern_version', 'regex', path)
                check_instance(settings_parsed['replace_version'], 'replace_version', 'string',
                               path)
                regex_version[catpkg] = (settings_parsed['pattern_version'],
                                         settings_parsed['replace_version'])
            if 'restrict_version' in settings_parsed:
                if settings_parsed.get('restrict_version').lower(
                ) != 'full' and settings_parsed.get('restrict_version').lower(
                ) != 'major' and settings_parsed.get('restrict_version').lower() != 'minor':
                    log.error('Invalid "restrict_version" in %s.', path)
                    continue
                restrict_version[catpkg] = settings_parsed['restrict_version'].lower()
            if 'sync_version' in settings_parsed:
                check_instance(settings_parsed['sync_version'], 'sync_version', 'string', path)
                sync_version[catpkg] = settings_parsed['sync_version']
            if 'stable_version' in settings_parsed:
                check_instance(settings_parsed['stable_version'], 'stable_version', 'regex', path)
                stable_version[catpkg] = settings_parsed['stable_version']

    return LivecheckSettings(branches, custom_livechecks, dotnet_projects, golang_packages,
                             type_packages, no_auto_update, sha_sources, transformations,
                             yarn_base_packages, yarn_packages, jetbrains_packages, keep_old,
                             gomodule_packages, gomodule_path, nodejs_packages, nodejs_path,
                             development, composer_packages, composer_path, regex_version,
                             restrict_version, sync_version, stable_version)


def check_instance(
        value: int | str | bool | list[str] | None,  # noqa: FBT001
        key: str,
        dtype: str,
        path: str | object,
        specific_value: bool | int | str | None = None) -> None:  # noqa: FBT001
    is_type = False
    if dtype == 'bool':
        is_type = isinstance(value, bool)
    elif dtype == 'int':
        is_type = isinstance(value, int)
    elif dtype == 'string':
        is_type = isinstance(value, str)
    elif dtype == 'none':
        is_type = value is None
    elif dtype == 'list':
        is_type = isinstance(value, list)
    elif dtype == 'url':
        if isinstance(value, str):
            parsed_url = urlparse(value)
            is_type = all([parsed_url.scheme, parsed_url.netloc])
    elif dtype == 'regex' and isinstance(value, str):
        try:
            re.compile(value)
            is_type = True
        except re.error:
            is_type = False

    if not is_type:
        log.error('value "%s" in key "%s" is not of type "%s" in file "%s.', value, key, dtype,
                  path)

    if specific_value is not None and value != specific_value:
        log.error('Value "%s" in key "%s" is not equal to "%s" in file "%s".', value, key,
                  specific_value, path)



"""Typing helpers."""
from __future__ import annotations

import requests

from .utils import TextDataResponse

__all__ = ('PropTuple', 'Response')

PropTuple = tuple[str, str, str, str, str, str, str]
"""A tuple for properties category, PN, PV, last version, top hash, hash date, and URL."""
Response = TextDataResponse | requests.Response
"""Special response type."""



# ruff: noqa: S108
from __future__ import annotations

from typing import TYPE_CHECKING

from livecheck.special.composer import (
    check_composer_requirements,
    remove_composer_url,
    update_composer_ebuild,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_remove_composer_url_calls_remove_url_ebuild(mocker: MockerFixture) -> None:
    mock_remove_url_ebuild = mocker.patch('livecheck.special.composer.remove_url_ebuild')
    mock_remove_url_ebuild.return_value = 'result'
    ebuild_content = 'SOME CONTENT'
    result = remove_composer_url(ebuild_content)
    mock_remove_url_ebuild.assert_called_once_with(ebuild_content, '-vendor.tar.xz')
    assert result == 'result'


def test_remove_composer_url_returns_expected_value(mocker: MockerFixture) -> None:
    expected = 'cleaned content'
    mocker.patch('livecheck.special.composer.remove_url_ebuild', return_value=expected)
    assert remove_composer_url('dummy') == expected


def test_update_composer_ebuild_no_composer_path(mocker: MockerFixture) -> None:
    mock_search_ebuild = mocker.patch('livecheck.special.composer.search_ebuild')
    mock_search_ebuild.return_value = (None, None)
    mock_sp_run = mocker.patch('livecheck.special.composer.sp.run')
    mock_build_compress = mocker.patch('livecheck.special.composer.build_compress')

    update_composer_ebuild('ebuild', 'path', {})
    mock_sp_run.assert_not_called()
    mock_build_compress.assert_not_called()


def test_update_composer_ebuild_success(mocker: MockerFixture) -> None:
    mock_search_ebuild = mocker.patch('livecheck.special.composer.search_ebuild')
    composer_path = '/tmp/composer'
    temp_dir = '/tmp/temp'
    mock_search_ebuild.return_value = (composer_path, temp_dir)
    mock_sp_run = mocker.patch('livecheck.special.composer.sp.run')
    mock_build_compress = mocker.patch('livecheck.special.composer.build_compress')

    fetchlist = {'foo': ('bar',)}
    update_composer_ebuild('ebuild', 'path', fetchlist)

    mock_sp_run.assert_called_once_with(('composer', '--no-interaction', '--no-scripts', 'install'),
                                        cwd=composer_path,
                                        check=True)
    mock_build_compress.assert_called_once_with(temp_dir, composer_path, 'vendor', '-vendor.tar.xz',
                                                fetchlist)


def test_update_composer_ebuild_sp_run_raises(mocker: MockerFixture) -> None:
    mock_search_ebuild = mocker.patch('livecheck.special.composer.search_ebuild')
    composer_path = '/tmp/composer'
    temp_dir = '/tmp/temp'
    mock_search_ebuild.return_value = (composer_path, temp_dir)
    mock_sp_run = mocker.patch('livecheck.special.composer.sp.run')
    mock_sp_run.side_effect = __import__('subprocess').CalledProcessError(1, 'composer')
    mock_build_compress = mocker.patch('livecheck.special.composer.build_compress')
    mock_log = mocker.patch('livecheck.special.composer.log')

    update_composer_ebuild('ebuild', 'path', {})

    mock_sp_run.assert_called_once()
    mock_build_compress.assert_not_called()
    assert mock_log.exception.called


def test_check_composer_requirements_returns_true_when_installed(mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.composer.check_program', return_value=True)
    mock_log = mocker.patch('livecheck.special.composer.log')
    assert check_composer_requirements() is True
    mock_check_program.assert_called_once_with('composer', ['--version'])
    mock_log.error.assert_not_called()


def test_check_composer_requirements_returns_false_when_not_installed(
        mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.composer.check_program',
                                      return_value=False)
    mock_log = mocker.patch('livecheck.special.composer.log')
    assert check_composer_requirements() is False
    mock_check_program.assert_called_once_with('composer', ['--version'])
    mock_log.error.assert_called_once_with('composer is not installed')



"""PECL functions."""
from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse

from defusedxml import ElementTree as ET  # noqa: N817
from livecheck.utils import assert_not_none, get_content
from livecheck.utils.portage import catpkg_catpkgsplit, get_last_version

if TYPE_CHECKING:
    from livecheck.settings import LivecheckSettings

__all__ = ('PECL_METADATA', 'get_latest_pecl_metadata', 'get_latest_pecl_package', 'is_pecl')

PECL_DOWNLOAD_URL = 'https://pecl.php.net/rest/r/%s/allreleases.xml'

PECL_METADATA = 'pecl'

NAMESPACE = '{http://pear.php.net/dtd/rest.allreleases}'


def get_latest_pecl_package(ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a PECL package."""
    _, _, program_name, _ = catpkg_catpkgsplit(ebuild)

    # Remove 'pecl-' prefix if present
    if program_name.startswith('pecl-'):
        program_name = program_name.replace('pecl-', '', 1)
    return get_latest_pecl_package2(program_name, ebuild, settings)


def get_latest_pecl_package2(program_name: str, ebuild: str, settings: LivecheckSettings) -> str:
    catpkg, _, _, _ = catpkg_catpkgsplit(ebuild)

    url = PECL_DOWNLOAD_URL % (program_name)

    if not (r := get_content(url)):
        return ''

    results: list[dict[str, str]] = []
    for release in ET.fromstring(r.text).findall(f'{NAMESPACE}r'):
        stability = release.find(f'{NAMESPACE}s')
        stability = assert_not_none(stability)
        if settings.is_devel(catpkg) or assert_not_none(stability.text) == 'stable':
            version = release.find(f'{NAMESPACE}v')
            version = assert_not_none(version)
            results.append({'tag': assert_not_none(version.text)})

    if last_version := get_last_version(results, '', ebuild, settings):
        return last_version['version']

    return ''


def is_pecl(url: str) -> bool:
    """Check if the URL is a PECL URL."""
    return urlparse(url).netloc == 'pecl.php.net'


def get_latest_pecl_metadata(remote: str, ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a PECL package."""
    return get_latest_pecl_package2(remote, ebuild, settings)



"""Special regular expression handling."""
from __future__ import annotations

from typing import TYPE_CHECKING
import logging
import re

from defusedxml import ElementTree as ET  # noqa: N817
from livecheck.constants import RSS_NS
from livecheck.utils import get_content, is_sha
from livecheck.utils.portage import catpkg_catpkgsplit, get_last_version

if TYPE_CHECKING:
    from livecheck.settings import LivecheckSettings

__all__ = ('get_latest_regex_package',)

logger = logging.getLogger(__name__)


def get_latest_regex_package(ebuild: str, url: str, regex: str,
                             settings: LivecheckSettings) -> tuple[str, str, str]:
    """Get the latest version of a package using a regular expression."""
    _, _, _, ebuild_version = catpkg_catpkgsplit(ebuild)

    if not (r := get_content(url)):
        return '', '', ''

    results: list[dict[str, str]] = []
    for result in re.findall(regex, r.text):
        if is_sha(result) and not results:
            logger.info('Found commit hash %s in %s.', result, url)
            hash_date = ''
            try:
                updated_el = ET.fromstring(r.text).find('entry/updated', RSS_NS)
            except ET.ParseError:
                logger.exception('Error parsing %s.', url)
                continue
            assert updated_el is not None
            assert updated_el.text is not None
            if re.search(r'(2[0-9]{7})', ebuild_version):
                hash_date = updated_el.text.split('T')[0].replace('-', '')
                logger.debug('Using updated date %s for commit %s.', hash_date, result)
            return result, hash_date, url
        results.append({'tag': result})

    if last_version := get_last_version(results, '', ebuild, settings):
        return last_version['version'], '', ''

    return '', '', ''



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, NoReturn
import operator
import re as real_re

from livecheck.utils.portage import (
    accept_version,
    catpkg_catpkgsplit,
    catpkgsplit2,
    compare_versions,
    digest_ebuild,
    fetch_ebuild,
    get_distdir,
    get_first_src_uri,
    get_highest_matches,
    get_last_version,
    get_repository_root_if_inside,
    is_version_development,
    mask_version,
    remove_initial_match,
    remove_leading_zeros,
    sanitize_version,
    unpack_ebuild,
)
import pytest

if TYPE_CHECKING:
    from collections.abc import Collection, Mapping

    from pytest_mock import MockerFixture


@pytest.mark.parametrize(('version', 'expected'), [
    (0, '0'),
    ('', ''),
    ('v1.2.3', '1.2.3'),
    ('s1.2.3', '1.2.3'),
    ('1.2.3', '1.2.3'),
    ('1.2.3a', '1.2.3a'),
    ('1.2.3-alpha', '1.2.3_alpha'),
    ('1.2.3-alpha0', '1.2.3_alpha'),
    ('1.2.3-beta1', '1.2.3_beta1'),
    ('1.2.3-beta.1', '1.2.3_beta1'),
    ('1.2.3-rc2', '1.2.3_rc2'),
    ('1.2.3_rc2', '1.2.3_rc2'),
    ('1.2.3 RC 2', '1.2.3_rc2'),
    ('1.2.3-pre', '1.2.3_pre'),
    ('1.2.3-dev', '1.2.3_beta'),
    ('1.2.3-test', '1.2.3_beta'),
    ('1.2.3_test', '1.2.3_beta'),
    ('1.2.3_test 1', '1.2.3_beta1'),
    ('1.2.3p', '1.2.3_p'),
    ('1.2.3-unknown', '1.2.3'),
    ('1.2.3-unknown1', '1.2.31'),
    ('1.2.3-1', '1.2.3.1'),
    ('1.2.3_1', '1.2.3.1'),
    ('1.2.3 1', '1.2.3.1'),
    ('1.2.3-rc', '1.2.3_rc'),
    ('1.2.3-rc-1', '1.2.3_rc1'),
    ('2022.12.26 2022-12-26 19:55', '2022.12.26'),
    ('2022_12_26 2022-12-26 19:55', '2022.12.26'),
    ('2022-12-26 2022-12-26 19:55', '2022.12.26'),
    ('dosbox 2022-12-26 2022-12-26 19:55', '2022.12.26'),
    ('samba-4.19.7', '4.19.7'),
    ('test-190', '190'),
    ('test 190', '190'),
    ('318.1', '318.1'),
    ('1.2.0.4068', '1.2.0.4068'),
    ('v2015-09-29-license-adobe', '2015.9.29'),
    ('dosbox-x-v2025.01.01', '2025.1.1'),
    ('v0.8.9p10', '0.8.9_p10'),
    ('build_420', '420'),
    ('glabels', ''),
    ('NewBuild25rc1', '25_rc1'),
    ('v1.12.post318', '1.12_p318'),
    ('1.002', '1.002'),
    ('1.3.0-build.4', '1.3.04'),
    ('0.2.tar.gz', '0.2'),
    ('0.8.1-pl5', '0.8.1_p5'),
    ('0.8 patchlevel   6', '0.8_p6'),
    ('0.0.8b2', '0.0.8_beta2'),
    ('0.0.8a5', '0.0.8_alpha5'),
    ('0.1.8b0', '0.1.8_beta'),
])
def test_sanitize_version(version: str, expected: str) -> None:
    assert sanitize_version(version) == expected


@pytest.mark.parametrize(('version', 'expected'), [
    ('2022.01.06', '2022.1.6'),
    ('24.01.12', '24.1.12'),
    ('0.0.3', '0.0.3'),
    ('1.0.3-r1', '1.0.3-r1'),
    ('2022-12-26', '2022-12-26'),
    ('1.0.3', '1.0.3'),
    ('1.2', '1.2'),
    ('1.222222222', '1.222222222'),
    ('1', '1'),
    ('0.1.2', '0.1.2'),
    ('24.01.02', '24.1.2'),
])
def test_remove_leading_zeros(version: str, expected: str) -> None:
    assert remove_leading_zeros(version) == expected


@pytest.mark.parametrize(('cp', 'version', 'restrict_version', 'expected'), [
    ('dev-util/foo', '1.2.3', 'major', 'dev-util/foo:1:'),
    ('dev-util/foo', '1.2.3', 'minor', 'dev-util/foo:1.2:'),
    ('dev-util/foo', '1.2.3', 'full', 'dev-util/foo'),
    ('dev-util/foo', '1.2.3', None, 'dev-util/foo'),
    ('dev-util/foo', '2.0', 'major', 'dev-util/foo:2:'),
    ('dev-util/foo', '2.0', 'minor', 'dev-util/foo:2.0:'),
    ('dev-util/foo', '2.0', 'full', 'dev-util/foo'),
    ('dev-util/foo', '2.0', '', 'dev-util/foo'),
])
def test_mask_version(cp: str, version: str, restrict_version: str | None, expected: str) -> None:
    assert mask_version(cp, version, restrict_version) == expected


def test_get_highest_matches_basic(mocker: MockerFixture) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    mocker.patch('livecheck.utils.portage.catpkg_catpkgsplit',
                 side_effect=lambda _: ('cat/pkg', 'cat', 'pkg', '1.2.3'))
    mock_p.xmatch.return_value = ['cat/pkg-1.2.3', 'cat/pkg-1.2.2']
    mock_p.findname2.return_value = ('cat/pkg-1.2.3', '/repo/root')

    names = ['cat/pkg']
    repo_root = Path('/repo/root')
    dummy_settings = mocker.Mock()
    dummy_settings.restrict_version = {}
    result = get_highest_matches(names, repo_root, dummy_settings)
    assert result == ['cat/pkg-1.2.3']


def test_get_highest_matches_no_matches(mocker: MockerFixture) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    mock_p.xmatch.return_value = []
    names = ['cat/pkg']
    repo_root = Path('/repo/root')
    dummy_settings = mocker.Mock()
    result = get_highest_matches(names, repo_root, dummy_settings)
    assert result == []


def test_get_highest_matches_invalid_package_structure(mocker: MockerFixture) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    mock_p.xmatch.return_value = ['invalid']

    def raise_value_error(x: Any) -> NoReturn:
        msg = 'bad structure'
        raise ValueError(msg)

    mocker.patch('livecheck.utils.portage.catpkg_catpkgsplit', side_effect=raise_value_error)
    names = ['cat/pkg']
    repo_root = Path('/repo/root')
    dummy_settings = mocker.Mock()
    result = get_highest_matches(names, repo_root, dummy_settings)
    assert result == []


def test_get_highest_matches_wrong_repo_root(mocker: MockerFixture) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    mock_p.xmatch.return_value = ['cat/pkg-1.2.3']
    mocker.patch('livecheck.utils.portage.catpkg_catpkgsplit',
                 return_value=('cat/pkg', 'cat', 'pkg', '1.2.3'))
    mock_p.findname2.return_value = ('cat/pkg-1.2.3', '/other/root')
    names = ['cat/pkg']
    repo_root = Path('/repo/root')
    dummy_settings = mocker.Mock()
    result = get_highest_matches(names, repo_root, dummy_settings)
    assert result == []


def test_get_highest_matches_ignores_9999(mocker: MockerFixture) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    mock_p.xmatch.return_value = ['cat/pkg-9999']
    mocker.patch('livecheck.utils.portage.catpkg_catpkgsplit',
                 return_value=('cat/pkg', 'cat', 'pkg', '9999'))
    mock_p.findname2.return_value = ('cat/pkg-9999', '/repo/root')
    names = ['cat/pkg']
    repo_root = Path('/repo/root')
    dummy_settings = mocker.Mock()
    result = get_highest_matches(names, repo_root, dummy_settings)
    assert result == []


def test_catpkgsplit2_valid_atom(mocker: MockerFixture) -> None:
    mock_catpkgsplit = mocker.patch('livecheck.utils.portage.catpkgsplit')
    mock_catpkgsplit.return_value = ('cat', 'pkg', '1.2.3', 'r0')
    catpkgsplit2.cache_clear()
    result = catpkgsplit2('cat/pkg-1.2.3')
    assert result == ('cat', 'pkg', '1.2.3', 'r0')


def test_catpkgsplit2_valid_atom_with_revision(mocker: MockerFixture) -> None:
    mock_catpkgsplit = mocker.patch('livecheck.utils.portage.catpkgsplit')
    mock_catpkgsplit.return_value = ('cat', 'pkg', '1.2.3', 'r1')
    catpkgsplit2.cache_clear()
    result = catpkgsplit2('cat/pkg-1.2.3-r1')
    assert result == ('cat', 'pkg', '1.2.3', 'r1')


def test_catpkgsplit2_none_returned(mocker: MockerFixture) -> None:
    mock_catpkgsplit = mocker.patch('livecheck.utils.portage.catpkgsplit')
    mock_catpkgsplit.return_value = None
    catpkgsplit2.cache_clear()
    with pytest.raises(ValueError, match='Invalid atom:'):
        catpkgsplit2('invalid-atom')


def test_catpkgsplit2_wrong_tuple_size(mocker: MockerFixture) -> None:
    mock_catpkgsplit = mocker.patch('livecheck.utils.portage.catpkgsplit')
    mock_catpkgsplit.return_value = ('cat', 'pkg', '1.2.3')  # Only 3 elements
    catpkgsplit2.cache_clear()
    with pytest.raises(ValueError, match='Invalid atom:'):
        catpkgsplit2('cat/pkg-1.2.3')


def test_catpkgsplit2_category_none(mocker: MockerFixture) -> None:
    mock_catpkgsplit = mocker.patch('livecheck.utils.portage.catpkgsplit')
    mock_catpkgsplit.return_value = (None, 'pkg', '1.2.3', 'r0')
    catpkgsplit2.cache_clear()
    result = catpkgsplit2('pkg-1.2.3')
    assert result == (None, 'pkg', '1.2.3', 'r0')


@pytest.mark.parametrize(
    ('atom', 'cat', 'pkg', 'ebuild_version', 'revision', 'expected'),
    [
        # No revision or r0
        ('cat/pkg-1.2.3', 'cat', 'pkg', '1.2.3', 'r0', ('cat/pkg', 'cat', 'pkg', '1.2.3')),
        # With revision r1
        ('cat/pkg-1.2.3-r1', 'cat', 'pkg', '1.2.3', 'r1', ('cat/pkg', 'cat', 'pkg', '1.2.3-r1')),
        # With revision r2
        ('cat/pkg-2.0.0-r2', 'cat', 'pkg', '2.0.0', 'r2', ('cat/pkg', 'cat', 'pkg', '2.0.0-r2')),
        # No revision, just version
        ('cat/pkg-0.1', 'cat', 'pkg', '0.1', '', ('cat/pkg', 'cat', 'pkg', '0.1')),
    ])
def test_catpkg_catpkgsplit_variants(mocker: MockerFixture, atom: str, cat: str, pkg: str,
                                     ebuild_version: str, revision: str,
                                     expected: tuple[str, str, str, str]) -> None:
    mock_catpkgsplit2 = mocker.patch('livecheck.utils.portage.catpkgsplit2')
    mock_catpkgsplit2.return_value = (cat, pkg, ebuild_version, revision)
    catpkg_catpkgsplit.cache_clear()
    result = catpkg_catpkgsplit(atom)
    assert result == expected


def test_catpkg_catpkgsplit_category_none(mocker: MockerFixture) -> None:
    mock_catpkgsplit2 = mocker.patch('livecheck.utils.portage.catpkgsplit2')
    mock_catpkgsplit2.return_value = (None, 'pkg', '1.2.3', 'r0')
    catpkg_catpkgsplit.cache_clear()
    with pytest.raises(AssertionError):
        catpkg_catpkgsplit('pkg-1.2.3')


def test_catpkg_catpkgsplit_revision_r0(mocker: MockerFixture) -> None:
    mock_catpkgsplit2 = mocker.patch('livecheck.utils.portage.catpkgsplit2')
    mock_catpkgsplit2.return_value = ('cat', 'pkg', '1.2.3', 'r0')
    catpkg_catpkgsplit.cache_clear()
    result = catpkg_catpkgsplit('cat/pkg-1.2.3')
    assert result == ('cat/pkg', 'cat', 'pkg', '1.2.3')


def test_catpkg_catpkgsplit_revision_non_r0(mocker: MockerFixture) -> None:
    mock_catpkgsplit2 = mocker.patch('livecheck.utils.portage.catpkgsplit2')
    mock_catpkgsplit2.return_value = ('cat', 'pkg', '1.2.3', 'r5')
    catpkg_catpkgsplit.cache_clear()
    result = catpkg_catpkgsplit('cat/pkg-1.2.3-r5')
    assert result == ('cat/pkg', 'cat', 'pkg', '1.2.3-r5')


@pytest.mark.parametrize(
    ('match', 'aux_get_return', 'expected'),
    [
        (
            'cat/pkg-1.2.3',
            ['http://example.com/src.tar.gz mirror://gentoo/src.tar.gz'],
            'http://example.com/src.tar.gz',
        ),
        (
            'cat/pkg-1.2.3',
            ['mirror://gentoo/src.tar.gz ftp://example.com/src.tar.gz'],
            'mirror://gentoo/src.tar.gz',
        ),
        (
            'cat/pkg-1.2.3',
            ['ftp://example.com/src.tar.gz'],
            'ftp://example.com/src.tar.gz',
        ),
        (
            'cat/pkg-1.2.3',
            ['not_a_uri something_else'],
            '',
        ),
        (
            'cat/pkg-1.2.3',
            [''],
            '',
        ),
    ],
)
def test_get_first_src_uri_basic(mocker: MockerFixture, match: str, aux_get_return: str,
                                 expected: str) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    mock_p.aux_get.return_value = aux_get_return
    result = get_first_src_uri(match)
    assert result == expected


def test_get_first_src_uri_with_search_dir(mocker: MockerFixture) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    mock_p.aux_get.return_value = ['https://example.com/foo.tar.gz']
    search_dir = Path('/some/dir')
    result = get_first_src_uri('cat/pkg-1.2.3', search_dir)
    mock_p.aux_get.assert_called_with('cat/pkg-1.2.3', ['SRC_URI'], mytree=str(search_dir))
    assert result == 'https://example.com/foo.tar.gz'


def test_get_first_src_uri_keyerror(mocker: MockerFixture) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    mock_p.aux_get.side_effect = KeyError('not found')
    result = get_first_src_uri('cat/pkg-1.2.3')
    assert not result


def test_get_first_src_uri_multiple_lines(mocker: MockerFixture) -> None:
    mock_p = mocker.patch('livecheck.utils.portage.P')
    # Simulate multiple lines, each line is split
    mock_p.aux_get.return_value = [
        'not_a_uri',
        'https://foo.com/bar.tar.gz',
        'mirror://gentoo/baz.tar.gz',
    ]
    result = get_first_src_uri('cat/pkg-1.2.3')
    assert result == 'https://foo.com/bar.tar.gz'


def test_get_repository_root_if_inside_inside_overlay(mocker: MockerFixture,
                                                      tmp_path: Path) -> None:
    # Setup fake repo structure
    repo_root = tmp_path / 'repo'
    repo_root.mkdir()
    subdir = repo_root / 'cat' / 'pkg'
    subdir.mkdir(parents=True)
    # Patch portage.config and settings
    mock_settings = {
        'PORTDIR_OVERLAY': str(repo_root),
        'PORTDIR': str(tmp_path / 'main-repo'),
    }
    mock_config = mocker.MagicMock()
    mock_config.__getitem__.side_effect = mock_settings.__getitem__
    mock_config.get.side_effect = mock_settings.get
    mocker.patch('livecheck.utils.portage.portage.config', return_value=mock_config)
    # Directory inside overlay
    result = get_repository_root_if_inside(subdir)
    assert result == (str(repo_root.resolve()), repo_root.name)


def test_get_repository_root_if_inside_inside_portdir(mocker: MockerFixture,
                                                      tmp_path: Path) -> None:
    portdir = tmp_path / 'main-repo'
    portdir.mkdir()
    subdir = portdir / 'foo'
    subdir.mkdir()
    mock_settings = {
        'PORTDIR_OVERLAY': str(tmp_path / 'overlay'),
        'PORTDIR': str(portdir),
    }
    mock_config = mocker.MagicMock()
    mock_config.__getitem__.side_effect = mock_settings.__getitem__
    mock_config.get.side_effect = mock_settings.get
    mocker.patch('livecheck.utils.portage.portage.config', return_value=mock_config)
    result = get_repository_root_if_inside(subdir)
    assert result == (str(portdir.resolve()), portdir.name)


def test_get_repository_root_if_inside_not_in_any_repo(mocker: MockerFixture,
                                                       tmp_path: Path) -> None:
    overlay = tmp_path / 'overlay'
    overlay.mkdir()
    portdir = tmp_path / 'main-repo'
    portdir.mkdir()
    outside = tmp_path / 'outside'
    outside.mkdir()
    mock_settings = {
        'PORTDIR_OVERLAY': str(overlay),
        'PORTDIR': str(portdir),
    }
    mock_config = mocker.MagicMock()
    mock_config.__getitem__.side_effect = mock_settings.__getitem__
    mock_config.get.side_effect = mock_settings.get
    mocker.patch('livecheck.utils.portage.portage.config', return_value=mock_config)
    result = get_repository_root_if_inside(outside)
    assert result == ('', '')


def test_get_repository_root_if_inside_local_path_exclusion(mocker: MockerFixture,
                                                            tmp_path: Path) -> None:
    overlay = tmp_path / 'overlay'
    overlay.mkdir()
    portdir = tmp_path / 'main-repo'
    portdir.mkdir()
    # Simulate a /local/ path
    local_dir = tmp_path / 'local' / 'repo'
    local_dir.mkdir(parents=True)
    mock_settings = {
        'PORTDIR_OVERLAY': str(overlay),
        'PORTDIR': str(portdir),
    }
    mock_config = mocker.MagicMock()
    mock_config.__getitem__.side_effect = mock_settings.__getitem__
    mock_config.get.side_effect = mock_settings.get
    mocker.patch('livecheck.utils.portage.portage.config', return_value=mock_config)
    result = get_repository_root_if_inside(local_dir)
    assert result == ('', '')


@pytest.mark.parametrize(('version', 'expected'), [
    ('1.2.3', False),
    ('1.2.3-alpha', True),
    ('1.2.3-beta', True),
    ('1.2.3-pre', True),
    ('1.2.3-dev', True),
    ('1.2.3-rc', True),
    ('1.2.3-ALPHA', True),
    ('1.2.3-BETA', True),
    ('1.2.3-PRE', True),
    ('1.2.3-DEV', True),
    ('1.2.3-RC', True),
    ('1.2.3-final', False),
    ('1.2.3-release', False),
    ('alpha', True),
    ('beta', True),
    ('pre', True),
    ('dev', True),
    ('rc', True),
    ('', False),
    ('1.2.3a', False),
    ('1.2.3b', False),
    ('1.2.3rc1', True),
    ('1.2.3dev1', True),
    ('1.2.3-pre1', True),
    ('1.2.3alpha', True),
    ('1.2.3beta', True),
    ('1.2.3rc', True),
    ('1.2.3dev', True),
    ('1.2.3-rc2', True),
    ('1.2.3-dev2', True),
    ('1.2.3-pre2', True),
])
def test_is_version_development(version: str, expected: bool) -> None:  # noqa: FBT001
    assert is_version_development(version) == expected


@pytest.mark.parametrize(('a', 'b', 'expected'), [
    ('foobar', 'foo', 'bar'),
    ('foo', 'foobar', ''),
    ('foobar', 'foobar', ''),
    ('foobar', '', 'foobar'),
    ('', 'foobar', ''),
    ('abcde', 'abc', 'de'),
    ('abc', 'abcde', ''),
    ('abc', 'abc', ''),
    ('abc', 'def', 'abc'),
    ('prefix_rest', 'prefix_', 'rest'),
    ('prefix_rest', 'prefix', '_rest'),
    ('prefix', 'prefix', ''),
    ('prefix', 'pre', 'fix'),
    ('', '', ''),
])
def test_remove_initial_match_various_cases(a: str, b: str, expected: str) -> None:
    assert remove_initial_match(a, b) == expected


@pytest.mark.parametrize(('old', 'new', 'vercmp_result', 'expected'), [
    ('1.2.3', '1.2.4', -1, True),
    ('1.2.3', '1.2.3', 0, False),
    ('1.2.4', '1.2.3', 1, False),
    ('2.0', '2.1', -1, True),
    ('2.1', '2.0', 1, False),
    ('1.0', '1.0.0', -1, True),
    ('1.0.0', '1.0', 1, False),
])
def test_compare_versions(mocker: MockerFixture, old: str, new: str, vercmp_result: int,
                          expected: bool) -> None:  # noqa: FBT001
    mock_vercmp = mocker.patch('livecheck.utils.portage.vercmp', return_value=vercmp_result)
    result = compare_versions(old, new)
    mock_vercmp.assert_called_once_with(old, new)
    assert result is expected


@pytest.mark.parametrize(
    ('distdir_value', 'expected'),
    [
        ('/custom/distdir', Path('/custom/distdir')),
        ('', Path('/var/cache/distfiles')),
        (None, Path('/var/cache/distfiles')),
    ],
)
def test_get_distdir(mocker: MockerFixture, distdir_value: str | None, expected: Path) -> None:
    mock_settings = {'DISTDIR': distdir_value} if distdir_value is not None else {}
    mock_config = mocker.MagicMock()
    mock_config.get.side_effect = mock_settings.get
    mocker.patch('livecheck.utils.portage.portage.config', return_value=mock_config)
    result = get_distdir()
    assert result == expected


@pytest.mark.parametrize(
    ('ebuild_path', 'doebuild_return', 'expected'),
    [
        ('/path/to/foo.ebuild', 0, True),
        ('/path/to/bar.ebuild', 1, False),
        ('/path/to/baz.ebuild', -1, False),
    ],
)
def test_fetch_ebuild_basic(mocker: MockerFixture, ebuild_path: str, doebuild_return: int,
                            expected: bool) -> None:  # noqa: FBT001
    mock_config = mocker.MagicMock()
    mock_portage = mocker.patch('livecheck.utils.portage.portage')
    mock_portage.config.return_value = mock_config
    mock_portage.doebuild.return_value = doebuild_return

    result = fetch_ebuild(ebuild_path)
    mock_portage.config.assert_called_once_with(clone=mock_portage.settings)
    mock_portage.doebuild.assert_called_once_with(ebuild_path,
                                                  'fetch',
                                                  settings=mock_config,
                                                  tree='porttree')
    assert result is expected


@pytest.mark.parametrize(
    ('ebuild_path', 'doebuild_return', 'expected'),
    [
        ('/path/to/foo.ebuild', 0, True),
        ('/path/to/bar.ebuild', 1, False),
        ('/path/to/baz.ebuild', -1, False),
    ],
)
def test_digest_ebuild_basic(mocker: MockerFixture, ebuild_path: str, doebuild_return: int,
                             expected: bool) -> None:  # noqa: FBT001
    mock_config = mocker.MagicMock()
    mock_portage = mocker.patch('livecheck.utils.portage.portage')
    mock_portage.config.return_value = mock_config
    mock_portage.doebuild.return_value = doebuild_return

    result = digest_ebuild(ebuild_path)
    mock_portage.config.assert_called_once_with(clone=mock_portage.settings)
    mock_portage.doebuild.assert_called_once_with(ebuild_path,
                                                  'digest',
                                                  settings=mock_config,
                                                  tree='porttree')
    assert result is expected


@pytest.mark.parametrize(
    ('clean_return', 'unpack_return', 'workdir_exists', 'workdir_is_dir', 'expected'),
    [
        # clean fails
        (1, 0, True, True, ''),
        # unpack fails
        (0, 1, True, True, ''),
        # workdir does not exist
        (0, 0, False, True, ''),
        # workdir exists but is not a dir
        (0, 0, True, False, ''),
        # workdir exists and is a dir
        (0, 0, True, True, '/some/workdir'),
    ],
)
def test_unpack_ebuild(
        mocker: MockerFixture,
        clean_return: int,
        unpack_return: int,
        workdir_exists: bool,  # noqa: FBT001
        workdir_is_dir: bool,  # noqa: FBT001
        expected: str) -> None:
    mock_config = mocker.MagicMock()
    mock_portage = mocker.patch('livecheck.utils.portage.portage')
    mock_portage.config.return_value = mock_config
    mock_portage.doebuild.side_effect = [clean_return, unpack_return]
    mock_config.__getitem__.return_value = '/some/workdir'

    mock_path = mocker.patch('livecheck.utils.portage.Path')
    mock_workdir = mock_path.return_value
    mock_workdir.exists.return_value = workdir_exists
    mock_workdir.is_dir.return_value = workdir_is_dir
    mock_workdir.__str__.return_value = '/some/workdir'

    result = unpack_ebuild('/path/to/foo.ebuild')

    assert result == expected
    mock_portage.config.assert_called_once_with(clone=mock_portage.settings)
    # Only call doebuild for clean and unpack if clean succeeds
    if clean_return != 0:
        assert mock_portage.doebuild.call_count == 1
        mock_portage.doebuild.assert_any_call('/path/to/foo.ebuild',
                                              'clean',
                                              settings=mock_config,
                                              tree='porttree')
    else:
        assert mock_portage.doebuild.call_count == 2
        mock_portage.doebuild.assert_any_call('/path/to/foo.ebuild',
                                              'clean',
                                              settings=mock_config,
                                              tree='porttree')
        mock_portage.doebuild.assert_any_call('/path/to/foo.ebuild',
                                              'unpack',
                                              settings=mock_config,
                                              tree='porttree')
    if clean_return == 0 and unpack_return == 0:
        mock_path.assert_called_with('/some/workdir')
        mock_workdir.exists.assert_called_once()
    else:
        mock_path.assert_not_called()


@pytest.mark.parametrize(
    ('results', 'repo', 'ebuild', 'settings_attrs', 'expected_version'),
    [
        # Basic: transformation and regex not set, sanitize_version used, valid version
        (
            [{
                'tag': '1.2.4'
            }, {
                'tag': '1.2.3'
            }],
            'repo',
            'cat/pkg-1.2.3',
            {
                'transformations': {},
                'regex_version': {},
                'restrict_version_process': '',
                'restrict_version': {},
                'stable_version': {},
                'is_devel': lambda _: False,
            },
            '1.2.4',
        ),
        # Transformation function present
        (
            [{
                'tag': 'v1.2.5'
            }, {
                'tag': 'v1.2.4'
            }],
            'repo',
            'cat/pkg-1.2.3',
            {
                'transformations': {
                    'cat/pkg': lambda tag: tag.lstrip('v')
                },
                'regex_version': {},
                'restrict_version_process': '',
                'restrict_version': {},
                'stable_version': {},
                'is_devel': lambda _: False,
            },
            '1.2.5',
        ),
        # Regex version present
        (
            [{
                'tag': 'foo-1.2.6'
            }, {
                'tag': 'foo-1.2.5'
            }],
            'repo',
            'cat/pkg-1.2.3',
            {
                'transformations': {},
                'regex_version': {
                    'cat/pkg': (r'foo-', '')
                },
                'restrict_version_process': '',
                'restrict_version': {},
                'stable_version': {},
                'is_devel': lambda _: False,
            },
            '1.2.6',
        ),
        # Version filtered out by restrict_version_process
        (
            [{
                'tag': '1.2.7'
            }, {
                'tag': '1.2.8'
            }],
            'repo',
            'cat/pkg-1.2.3',
            {
                'transformations': {},
                'regex_version': {},
                'restrict_version_process': '2.8',  # No version startswith 2.8
                'restrict_version': {},
                'stable_version': {},
                'is_devel': lambda _: False,
            },
            None,
        ),
        # Version filtered out by catpkg_catpkgsplit ValueError
        (
            [{
                'tag': 'bad_ver'
            }],
            'repo',
            'cat/pkg-1.2.3',
            {
                'transformations': {},
                'regex_version': {},
                'restrict_version_process': '',
                'restrict_version': {},
                'stable_version': {},
                'is_devel': lambda _: False,
            },
            None,
        ),
        # Accept version returns False
        (
            [{
                'tag': '1.2.9'
            }],
            'repo',
            'cat/pkg-1.2.3',
            {
                'transformations': {},
                'regex_version': {},
                'restrict_version_process': '',
                'restrict_version': {},
                'stable_version': {},
                'is_devel': lambda _: False,
                'accept_version': False,
            },
            None,
        ),
        # Ebuild version has more than one dot, but tag version has none (should skip)
        (
            [{
                'tag': 'foo'
            }],
            'repo',
            'cat/pkg-1.2.3.4',
            {
                'transformations': {},
                'regex_version': {},
                'restrict_version_process': '',
                'restrict_version': {},
                'stable_version': {},
                'is_devel': lambda _: False,
            },
            None,
        ),
    ])
def test_get_last_version_cases(mocker: MockerFixture, results: Collection[Mapping[str, str]],
                                repo: str, ebuild: str, settings_attrs: Mapping[str, Any],
                                expected_version: str | None) -> None:
    # Patch catpkg_catpkgsplit to handle ebuild and tag versions
    def fake_catpkg_catpkgsplit(atom: str) -> tuple[str, str, str, str]:
        if atom == ebuild:
            return ('cat/pkg', 'cat', 'pkg', '1.2.3')
        if atom.startswith('cat/pkg-'):
            version = atom.split('-', 1)[1]
            if version == 'bad_ver':
                msg = 'bad version'
                raise ValueError(msg)
            return ('cat/pkg', 'cat', 'pkg', version)
        msg = 'bad atom'
        raise ValueError(msg)

    mocker.patch('livecheck.utils.portage.catpkg_catpkgsplit', side_effect=fake_catpkg_catpkgsplit)
    mocker.patch('livecheck.utils.portage.sanitize_version', side_effect=lambda v, _: v)
    mocker.patch('livecheck.utils.portage.compare_versions', side_effect=operator.lt)

    # Patch accept_version if requested
    if settings_attrs.get('accept_version', True) is False:
        mocker.patch('livecheck.utils.portage.accept_version', return_value=False)
    else:
        mocker.patch('livecheck.utils.portage.accept_version', return_value=True)

    # Build dummy settings
    dummy_settings = mocker.Mock()
    for k, v in settings_attrs.items():
        setattr(dummy_settings, k, v)

    result = get_last_version(results, repo, ebuild, dummy_settings)
    if expected_version is None:
        assert result == {}
    else:
        assert result['version'] == expected_version


def test_get_last_version_sanitize_version_returns_empty(mocker: MockerFixture) -> None:
    # Patch catpkg_catpkgsplit to handle ebuild and tag versions
    def fake_catpkg_catpkgsplit(atom: str) -> tuple[str, str, str, str]:
        if atom == 'cat/pkg-1.2.3':
            return ('cat/pkg', 'cat', 'pkg', '1.2.3')
        msg = 'bad atom'
        raise ValueError(msg)

    mocker.patch('livecheck.utils.portage.catpkg_catpkgsplit', side_effect=fake_catpkg_catpkgsplit)
    mocker.patch('livecheck.utils.portage.sanitize_version', return_value='')

    results = [{'tag': '1.2.3'}]
    repo = 'repo'
    ebuild = 'cat/pkg-1.2.3'
    dummy_settings = mocker.Mock()
    dummy_settings.transformations = {}
    dummy_settings.regex_version = {}
    dummy_settings.restrict_version_process = ''
    dummy_settings.restrict_version = {}
    dummy_settings.stable_version = {}
    dummy_settings.is_devel = lambda _: False

    result = get_last_version(results, repo, ebuild, dummy_settings)
    assert result == {}


def test_get_last_version_catpkg_catpkgsplit_raises_value_error(mocker: MockerFixture) -> None:
    # Patch catpkg_catpkgsplit to raise ValueError
    mocker.patch('livecheck.utils.portage.catpkg_catpkgsplit',
                 side_effect=[('cat/pkg', 'cat', 'pkg', '1.2.3'),
                              ValueError('bad atom')])

    results = [{'tag': '1.2.3'}]
    repo = 'repo'
    ebuild = 'cat/pkg-1.2.3'
    dummy_settings = mocker.Mock()
    dummy_settings.transformations = {}
    dummy_settings.regex_version = {}
    dummy_settings.restrict_version_process = ''
    dummy_settings.restrict_version = {}
    dummy_settings.stable_version = {}
    dummy_settings.is_devel = lambda _: False

    result = get_last_version(results, repo, ebuild, dummy_settings)
    assert result == {}


@pytest.mark.parametrize(
    ('ebuild_version', 'version', 'catpkg', 'settings_attrs', 'expected'),
    [
        # ebuild_version is development, should return True
        ('1.2.3-alpha', '1.2.4', 'cat/pkg', {
            'stable_version': {},
            'is_devel': lambda _: False,
        }, True),
        # settings.is_devel returns True, should return True
        ('1.2.3', '1.2.4', 'cat/pkg', {
            'stable_version': {},
            'is_devel': lambda _: True,
        }, True),
        # stable_version regex matches version, should return True
        ('1.2.3', '2.0.0', 'cat/pkg', {
            'stable_version': {
                'cat/pkg': r'^2\..*'
            },
            'is_devel': lambda _: False,
        }, True),
        # version is development, should return False
        ('1.2.3', '1.2.4-beta', 'cat/pkg', {
            'stable_version': {},
            'is_devel': lambda _: False,
        }, False),
        # stable_version regex does not match version, should return False
        ('1.2.3', '3.0.0', 'cat/pkg', {
            'stable_version': {
                'cat/pkg': r'^2\..*'
            },
            'is_devel': lambda _: False,
        }, False),
        # Neither development nor stable, should return True
        ('1.2.3', '1.2.4', 'cat/pkg', {
            'stable_version': {},
            'is_devel': lambda _: False,
        }, True),
        # Both ebuild_version and version are not development, but stable_version is set and matches
        ('1.2.3', '2.1.0', 'cat/pkg', {
            'stable_version': {
                'cat/pkg': r'^2\..*'
            },
            'is_devel': lambda _: False,
        }, True),
        # Both ebuild_version and version are not development, but stable_version is set and does
        # not match.
        ('1.2.3', '3.1.0', 'cat/pkg', {
            'stable_version': {
                'cat/pkg': r'^2\..*'
            },
            'is_devel': lambda _: False,
        }, False),
    ])
def test_accept_version_cases(mocker: MockerFixture, ebuild_version: str, version: str, catpkg: str,
                              settings_attrs: Mapping[str,
                                                      Any], expected: bool) -> None:  # noqa: FBT001
    # Patch is_version_development to match the real implementation
    mocker.patch(
        'livecheck.utils.portage.is_version_development',
        side_effect=lambda v: 'alpha' in v or 'beta' in v or 'dev' in v or 'rc' in v or 'pre' in v)
    # Patch re.match to use the real re.match
    mocker.patch('livecheck.utils.portage.re', real_re)
    # Build dummy settings
    dummy_settings = mocker.Mock()
    for k, v in settings_attrs.items():
        setattr(dummy_settings, k, v)
    result = accept_version(ebuild_version, version, catpkg, dummy_settings)
    assert result is expected



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from livecheck.special.metacpan import (
    extract_perl_package,
    get_latest_metacpan_metadata,
    get_latest_metacpan_package,
    is_metacpan,
)
import pytest

if TYPE_CHECKING:
    from collections.abc import Collection, Mapping, Sequence

    from pytest_mock import MockerFixture

test_cases = {
    'valid_url': {
        'url': 'https://metacpan.org/release/Foo-Bar-1.23',
        'expected': 'Foo-Bar',
        'is_metacpan': True
    },
    'valid_url_with_version': {
        'url': 'https://metacpan.org/release/Foo-Bar-1.23.tar.gz',
        'expected': 'Foo-Bar',
        'is_metacpan': True
    },
    'invalid_url': {
        'url': 'https://example.com/release/Foo-Bar-1.23',
        'expected': '',
        'is_metacpan': False
    },
    'invalid_url_no_version': {
        'url': 'https://metacpan.org/release/Foo-Bar',
        'expected': '',
        'is_metacpan': False
    },
    'cpan_url': {
        'url': 'mirror://cpan/release/Foo-Bar-1.23',
        'expected': 'Foo-Bar',
        'is_metacpan': True
    },
    'cpan_url_no_version': {
        'url': 'mirror://cpan/release/Foo-Bar',
        'expected': '',
        'is_metacpan': False
    },
    'net-mail/grepmail': {
        'url': 'mirror://cpan/authors/id/D/DC/DCOPPIT/grepmail-5.3111.tar.gz',
        'expected': 'grepmail',
        'is_metacpan': True
    },
}


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_extract_perl_package(test_case: dict[str, Any]) -> None:
    assert extract_perl_package(test_case['url']) == test_case['expected']


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_is_metacpan(test_case: dict[str, Any]) -> None:
    assert is_metacpan(test_case['url']) == test_case['is_metacpan']


@pytest.mark.parametrize(
    ('url', 'ebuild', 'api_hits', 'api_release', 'expected_version'),
    [
        (
            'https://metacpan.org/release/Foo-Bar-1.23',
            'foo-bar.ebuild',
            [{
                '_source': {
                    'version': '1.23'
                }
            }, {
                '_source': {
                    'version': '1.22'
                }
            }],
            {
                'version': '1.23'
            },
            '1.23',
        ),
        (
            'https://metacpan.org/release/Foo-Bar-2.00',
            'foo-bar.ebuild',
            [],
            {
                'version': '2.00'
            },
            '2.00',
        ),
        (
            'https://metacpan.org/release/Foo-Bar-3.00',
            'foo-bar.ebuild',
            [],
            {},
            '',
        ),
    ],
)
def test_get_latest_metacpan_package(mocker: MockerFixture, url: str, ebuild: str,
                                     api_hits: Collection[Any], api_release: Mapping[str, Any],
                                     expected_version: str) -> None:
    def fake_get_content(url_arg: str) -> Any:
        class FakeResponse:
            def json(self) -> Any:
                if 'release/_search' in url_arg:
                    return {'hits': {'hits': api_hits}}
                return api_release

        if not api_hits and not api_release:
            return None
        return FakeResponse()

    def fake_get_last_version(results: Sequence[Any], package_name: str, ebuild: str,
                              settings: Any) -> dict[str, str] | None:
        if not results or (not results[-1].get('tag') and not results[-1].get('version')):
            return None
        last = results[-1]
        return {'version': last.get('tag') or last.get('version')}

    mocker.patch('livecheck.special.metacpan.get_content', side_effect=fake_get_content)
    mocker.patch('livecheck.special.metacpan.get_last_version', side_effect=fake_get_last_version)
    settings = mocker.Mock()
    result = get_latest_metacpan_package(url, ebuild, settings)
    assert result == expected_version


@pytest.mark.parametrize(
    ('remote', 'ebuild', 'api_hits', 'api_release', 'expected_version'),
    [
        (
            'Foo-Bar',
            'foo-bar.ebuild',
            [{
                '_source': {
                    'version': '1.50'
                }
            }, {
                '_source': {
                    'version': '1.40'
                }
            }],
            {
                'version': '1.50'
            },
            '1.50',
        ),
        (
            'Baz-Quux',
            'baz-quux.ebuild',
            [],
            {
                'version': '2.10'
            },
            '2.10',
        ),
        (
            'NoVersion',
            'no-version.ebuild',
            [],
            {},
            '',
        ),
    ],
)
def test_get_latest_metacpan_metadata(mocker: MockerFixture, remote: str, ebuild: str,
                                      api_hits: Collection[Any], api_release: Mapping[str, Any],
                                      expected_version: str) -> None:
    def fake_get_content(url_arg: str) -> Any:
        class FakeResponse:
            def json(self) -> Any:
                if 'release/_search' in url_arg:
                    return {'hits': {'hits': api_hits}}
                return api_release

        if not api_hits and not api_release:
            return None
        return FakeResponse()

    def fake_get_last_version(results: Sequence[Any], package_name: str, ebuild: str,
                              settings: Any) -> dict[str, str] | None:
        if not results or (not results[-1].get('tag') and not results[-1].get('version')):
            return None
        last = results[-1]
        return {'version': last.get('tag') or last.get('version')}

    mocker.patch('livecheck.special.metacpan.get_content', side_effect=fake_get_content)
    mocker.patch('livecheck.special.metacpan.get_last_version', side_effect=fake_get_last_version)
    settings = mocker.Mock()
    result = get_latest_metacpan_metadata(remote, ebuild, settings)
    assert result == expected_version



"""RubyGems functions."""
from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse

from livecheck.utils import get_content
from livecheck.utils.portage import catpkg_catpkgsplit, get_last_version

if TYPE_CHECKING:
    from livecheck.settings import LivecheckSettings

__all__ = ('RUBYGEMS_METADATA', 'get_latest_rubygems_metadata', 'get_latest_rubygems_package',
           'is_rubygems')

RUBYGEMS_DOWNLOAD_URL = 'https://rubygems.org/api/v1/versions/%s.json'
RUBYGEMS_METADATA = 'rubygems'


def get_latest_rubygems_package(ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a RubyGems package."""
    _, _, gem_name, _ = catpkg_catpkgsplit(ebuild)
    return get_latest_rubygems_package2(gem_name, ebuild, settings)


def get_latest_rubygems_package2(gem_name: str, ebuild: str, settings: LivecheckSettings) -> str:
    catpkg, _, _, _ = catpkg_catpkgsplit(ebuild)
    url = RUBYGEMS_DOWNLOAD_URL % (gem_name)

    if not (response := get_content(url)):
        return ''

    results: list[dict[str, str]] = [
        {
            'tag': release.get('number', '')
        } for release in response.json()
        if settings.is_devel(catpkg) or not release.get('prerelease', False)
    ]

    if last_version := get_last_version(results, gem_name, ebuild, settings):
        return last_version['version']

    return ''


def is_rubygems(url: str) -> bool:
    """Check if the URL is a RubyGems URL."""
    return urlparse(url).netloc == 'rubygems.org'


def get_latest_rubygems_metadata(remote: str, ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a RubyGems package."""
    return get_latest_rubygems_package2(remote, ebuild, settings)



# ruff: noqa: S108
from __future__ import annotations

from typing import TYPE_CHECKING
import subprocess as sp

from livecheck.special.nodejs import (
    check_nodejs_requirements,
    remove_nodejs_url,
    update_nodejs_ebuild,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_remove_nodejs_url_removes_node_modules_line() -> None:
    ebuild_content = """
SRC_URI="https://example.com/foo.tar.gz
https://example.com/foo-node_modules.tar.xz"
"""
    result = remove_nodejs_url(ebuild_content)
    assert '-node_modules.tar.xz' not in result
    assert 'foo.tar.gz' in result


def test_remove_nodejs_url_no_node_modules_line() -> None:
    ebuild_content = """
SRC_URI="https://example.com/foo.tar.gz"
"""
    result = remove_nodejs_url(ebuild_content)
    assert result == ebuild_content


def test_remove_nodejs_url_multiple_node_modules_lines() -> None:
    ebuild_content = """
SRC_URI="https://example.com/foo-node_modules.tar.xz
https://example.com/bar-node_modules.tar.xz
https://example.com/baz.tar.gz"
"""
    result = remove_nodejs_url(ebuild_content)
    assert '-node_modules.tar.xz' not in result
    assert 'baz.tar.gz' in result


def test_update_nodejs_ebuild_success(mocker: MockerFixture) -> None:
    mock_search_ebuild = mocker.patch('livecheck.special.nodejs.search_ebuild')
    mock_run = mocker.patch('livecheck.special.nodejs.sp.run')
    mock_build_compress = mocker.patch('livecheck.special.nodejs.build_compress')

    ebuild = 'dummy.ebuild'
    path = '/some/path'
    fetchlist = {'foo': ('bar',)}
    package_path = '/tmp/pkg'
    temp_dir = '/tmp/tmpdir'

    mock_search_ebuild.return_value = (package_path, temp_dir)

    update_nodejs_ebuild(ebuild, path, fetchlist)

    mock_search_ebuild.assert_called_once_with(ebuild, 'package.json', path)
    mock_run.assert_called_once_with(('npm', 'install', '--audit false', '--color false',
                                      '--progress false', '--ignore-scripts'),
                                     cwd=package_path,
                                     check=True)
    mock_build_compress.assert_called_once_with(temp_dir, package_path, 'node_modules',
                                                '-node_modules.tar.xz', fetchlist)


def test_update_nodejs_ebuild_no_package_path(mocker: MockerFixture) -> None:
    mock_search_ebuild = mocker.patch('livecheck.special.nodejs.search_ebuild')
    mock_run = mocker.patch('livecheck.special.nodejs.sp.run')
    mock_build_compress = mocker.patch('livecheck.special.nodejs.build_compress')

    mock_search_ebuild.return_value = (None, None)

    update_nodejs_ebuild('dummy.ebuild', None, {})

    mock_search_ebuild.assert_called_once()
    mock_run.assert_not_called()
    mock_build_compress.assert_not_called()


def test_update_nodejs_ebuild_npm_install_fails(mocker: MockerFixture) -> None:
    mock_search_ebuild = mocker.patch('livecheck.special.nodejs.search_ebuild')
    mock_run = mocker.patch('livecheck.special.nodejs.sp.run')
    mock_build_compress = mocker.patch('livecheck.special.nodejs.build_compress')
    mock_logger = mocker.patch('livecheck.special.nodejs.logger')

    package_path = '/tmp/pkg'
    temp_dir = '/tmp/tmpdir'
    mock_search_ebuild.return_value = (package_path, temp_dir)
    mock_run.side_effect = sp.CalledProcessError(1, 'npm')

    update_nodejs_ebuild('dummy.ebuild', None, {})

    mock_run.assert_called_once()
    mock_build_compress.assert_not_called()
    assert mock_logger.exception.call_count == 1


def test_check_nodejs_requirements_success(mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.nodejs.check_program')
    mock_logger = mocker.patch('livecheck.special.nodejs.logger')
    mock_check_program.return_value = True

    result = check_nodejs_requirements()

    mock_check_program.assert_called_once_with('npm', ['--version'])
    assert result is True
    mock_logger.error.assert_not_called()


def test_check_nodejs_requirements_failure(mocker: MockerFixture) -> None:
    mock_check_program = mocker.patch('livecheck.special.nodejs.check_program')
    mock_logger = mocker.patch('livecheck.special.nodejs.logger')
    mock_check_program.return_value = False

    result = check_nodejs_requirements()

    mock_check_program.assert_called_once_with('npm', ['--version'])
    assert result is False
    mock_logger.error.assert_called_once_with('npm is not installed')



# ruff: noqa: FBT001
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from livecheck.special.bitbucket import (
    extract_workspace_and_repository,
    get_latest_bitbucket,
    get_latest_bitbucket_metadata,
    get_latest_bitbucket_package,
    is_bitbucket,
)
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

test_cases = {
    'atlassian': {
        'url': 'https://bitbucket.org/atlassian/python-bitbucket',
        'expected': ('atlassian', 'python-bitbucket'),
        'is_bitbucket': True
    },
    'atlassian_master': {
        'url': 'https://bitbucket.org/atlassian/python-bitbucket/src/master/',
        'expected': ('atlassian', 'python-bitbucket'),
        'is_bitbucket': True
    },
    'bad_url': {
        'url': 'https://bitbucket.org/',
        'expected': ('', ''),
        'is_bitbucket': False
    },
    'bad_url2': {
        'url': 'https://github.io/username/repo.git',
        'expected': ('', ''),
        'is_bitbucket': False
    },
    'bad_url3': {
        'url': 'https://bitbucket.org/atlassian/',
        'expected': ('', ''),
        'is_bitbucket': False
    },
    'bitbucket_repo_git': {
        'url': 'https://bitbucket.org/atlassian/python-bitbucket.git',
        'expected': ('atlassian', 'python-bitbucket'),
        'is_bitbucket': True
    },
    'media-plugins/vdr-dvbhddevice': {
        'url': 'https://bitbucket.org/powARman/dvbhddevice/get/20170225.tar.bz2',
        'expected': ('powARman', 'dvbhddevice'),
        'is_bitbucket': True
    },
}


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_extract_owner_repo(test_case: dict[str, Any]) -> None:
    assert extract_workspace_and_repository(test_case['url']) == test_case['expected']


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_is_github(test_case: dict[str, Any]) -> None:
    assert is_bitbucket(test_case['url']) == test_case['is_bitbucket']


def make_mock_response(json_data: Any, *, ok: bool = True) -> Any:
    @dataclass
    class MockResponse:
        ok: bool = True

        def json(self) -> Any:
            return json_data

    return MockResponse(ok=ok)


def test_get_latest_bitbucket_package_tags_only(mocker: MockerFixture) -> None:
    # Simulate tags response with two tags
    tags_json = {
        'values': [
            {
                'name': 'v1.0.0',
                'target': {
                    'hash': 'abc123'
                }
            },
            {
                'name': 'v2.0.0',
                'target': {
                    'hash': 'def456'
                }
            },
        ]
    }
    downloads_json: dict[str, Any] = {'values': []}
    mock_get_content = mocker.patch(
        'livecheck.special.bitbucket.get_content',
        side_effect=[
            make_mock_response(tags_json),  # tags
            make_mock_response(downloads_json),  # downloads
        ])
    mock_get_last_version = mocker.patch('livecheck.special.bitbucket.get_last_version',
                                         return_value={
                                             'version': 'v2.0.0',
                                             'id': 'def456'
                                         })
    url = 'https://bitbucket.org/atlassian/python-bitbucket'
    cpv = 'dev-python/bitbucket-2.0.0'
    version, commit = get_latest_bitbucket_package(url, cpv, mocker.Mock())
    assert version == 'v2.0.0'
    assert commit == 'def456'
    assert mock_get_content.call_count == 2
    assert mock_get_last_version.called


def test_get_latest_bitbucket_package_downloads(mocker: MockerFixture) -> None:
    tags_json: dict[str, Any] = {'values': []}
    downloads_json = {
        'values': [
            {
                'name': 'python-bitbucket-3.0.0.tar.gz'
            },
            {
                'name': 'python-bitbucket-2.5.0.zip'
            },
        ],
        'next': None
    }
    mock_get_content = mocker.patch(
        'livecheck.special.bitbucket.get_content',
        side_effect=[
            make_mock_response(tags_json),  # tags
            make_mock_response(downloads_json),  # downloads
        ])
    mock_get_last_version = mocker.patch('livecheck.special.bitbucket.get_last_version',
                                         return_value={
                                             'version': '3.0.0',
                                             'id': ''
                                         })
    url = 'https://bitbucket.org/atlassian/python-bitbucket'
    cpv = 'dev-python/bitbucket-3.0.0'
    version, commit = get_latest_bitbucket_package(url, cpv, mocker.Mock())
    assert version == '3.0.0'
    assert not commit
    assert mock_get_content.call_count == 2
    assert mock_get_last_version.called


def test_get_latest_bitbucket_package_response_not_ok(mocker: MockerFixture) -> None:
    tags_json: dict[str, Any] = {'values': []}
    mock_get_content = mocker.patch(
        'livecheck.special.bitbucket.get_content',
        side_effect=[
            make_mock_response(tags_json),  # tags
            make_mock_response({}, ok=False),  # downloads
        ])
    mock_get_last_version = mocker.patch('livecheck.special.bitbucket.get_last_version',
                                         return_value={
                                             'version': '3.0.0',
                                             'id': ''
                                         })
    url = 'https://bitbucket.org/atlassian/python-bitbucket'
    cpv = 'dev-python/bitbucket-3.0.0'
    version, commit = get_latest_bitbucket_package(url, cpv, mocker.Mock())
    assert version == '3.0.0'
    assert not commit
    assert mock_get_content.call_count == 2
    assert mock_get_last_version.called


def test_get_latest_bitbucket_package_no_tags_no_downloads(mocker: MockerFixture) -> None:
    tags_json: dict[str, Any] = {'values': []}
    downloads_json: dict[str, Any] = {'values': [], 'next': None}
    mock_get_content = mocker.patch(
        'livecheck.special.bitbucket.get_content',
        side_effect=[
            make_mock_response(tags_json),  # tags
            make_mock_response(downloads_json),  # downloads
        ])
    mock_get_last_version = mocker.patch('livecheck.special.bitbucket.get_last_version',
                                         return_value=None)
    url = 'https://bitbucket.org/atlassian/python-bitbucket'
    cpv = 'dev-python/bitbucket-0.0.1'
    version, commit = get_latest_bitbucket_package(url, cpv, mocker.Mock())
    assert not version
    assert not commit
    assert mock_get_content.call_count == 2
    assert mock_get_last_version.called


def test_get_latest_bitbucket_package_get_content_fails(mocker: MockerFixture) -> None:
    mock_get_content = mocker.patch('livecheck.special.bitbucket.get_content', return_value=None)
    url = 'https://bitbucket.org/atlassian/python-bitbucket'
    cpv = 'dev-python/bitbucket-0.0.1'
    version, commit = get_latest_bitbucket_package(url, cpv, mocker.Mock())
    assert not version
    assert not commit
    mock_get_content.assert_called_once()


@pytest.mark.parametrize(
    ('url', 'cpv', 'force_sha', 'is_sha_return', 'expected_version', 'expected_top_hash',
     'expected_hash_date'),
    [
        (
            'https://bitbucket.org/atlassian/python-bitbucket',
            'dev-python/bitbucket-2.0.0',
            False,
            False,
            'v2.0.0',
            '',
            '',
        ),
        (
            'https://bitbucket.org/atlassian/python-bitbucket',
            'dev-python/bitbucket-2.0.0',
            True,
            False,
            'v2.0.0',
            'def456',
            '',
        ),
        (
            'https://bitbucket.org/atlassian/python-bitbucket/commit/abc123',
            'dev-python/bitbucket-2.0.0',
            False,
            True,
            '',
            '',
            '',
        ),
    ],
)
def test_get_latest_bitbucket(mocker: MockerFixture, url: str, cpv: str, force_sha: bool,
                              is_sha_return: bool, expected_version: str, expected_top_hash: str,
                              expected_hash_date: str) -> None:
    # Patch is_sha to control commit detection
    mocker.patch('livecheck.special.bitbucket.is_sha', return_value=is_sha_return)
    mock_log_unhandled_commit = mocker.patch('livecheck.special.bitbucket.log_unhandled_commit')
    mock_get_latest_bitbucket_package = mocker.patch(
        'livecheck.special.bitbucket.get_latest_bitbucket_package',
        return_value=('v2.0.0', 'def456'),
    )

    version, top_hash, hash_date = get_latest_bitbucket(url,
                                                        cpv,
                                                        mocker.Mock(),
                                                        force_sha=force_sha)

    if is_sha_return:
        mock_log_unhandled_commit.assert_called_once_with(cpv, url)
        mock_get_latest_bitbucket_package.assert_not_called()
        assert not version
        assert not top_hash
        assert hash_date == expected_hash_date
    else:
        mock_log_unhandled_commit.assert_not_called()
        mock_get_latest_bitbucket_package.assert_called_once_with(url, cpv, mocker.ANY)
        assert version == expected_version
        assert top_hash == expected_top_hash
        assert hash_date == expected_hash_date


def test_get_latest_bitbucket_metadata(mocker: MockerFixture) -> None:
    mock_get_latest_bitbucket_package = mocker.patch(
        'livecheck.special.bitbucket.get_latest_bitbucket_package',
        return_value=('v2.0.0', 'def456'),
    )
    remote = 'atlassian/python-bitbucket'
    cpv = 'dev-python/bitbucket-2.0.0'
    version, commit = get_latest_bitbucket_metadata(remote, cpv, mocker.Mock())
    assert version == 'v2.0.0'
    assert commit == 'def456'
    mock_get_latest_bitbucket_package.assert_called_once_with(f'https://bitbucket.org/{remote}',
                                                              cpv, mocker.ANY)



"""Golang (go-module) utilities."""
from __future__ import annotations

from typing import TYPE_CHECKING
import logging
import subprocess as sp

from livecheck.utils import check_program

from .utils import build_compress, remove_url_ebuild, search_ebuild

if TYPE_CHECKING:
    from collections.abc import Mapping

__all__ = ('check_gomodule_requirements', 'remove_gomodule_url', 'update_gomodule_ebuild')

logger = logging.getLogger(__name__)


def remove_gomodule_url(ebuild_content: str) -> str:
    """Remove ``-vendor.tar.xz`` line from ebuild."""
    return remove_url_ebuild(ebuild_content, '-vendor.tar.xz')


def update_gomodule_ebuild(ebuild: str, path: str | None,
                           fetchlist: Mapping[str, tuple[str, ...]]) -> None:
    """Update a Go module-based ebuild."""
    go_mod_path, temp_dir = search_ebuild(ebuild, 'go.mod', path)
    if not go_mod_path:
        return

    try:
        sp.run(('go', 'mod', 'vendor'), cwd=go_mod_path, check=True)
    except sp.CalledProcessError:
        logger.exception("Error running 'go mod vendor'.")
        return

    build_compress(temp_dir, go_mod_path, 'vendor', '-vendor.tar.xz', fetchlist)


def check_gomodule_requirements() -> bool:
    """Check if Go is installed."""
    if not check_program('go', ['version']):
        logger.error('go is not installed')
        return False
    return True



"""Golang functions."""
from __future__ import annotations

from pathlib import Path
import contextlib
import logging
import re

from livecheck.utils import get_content

from .utils import EbuildTempFile

__all__ = ('update_go_ebuild',)

logger = logging.getLogger(__name__)


class InvalidGoSumURITemplate(ValueError):
    """Raised when the Go sum URI template is invalid."""
    def __init__(self) -> None:
        super().__init__('URI template missing @PV@ or @SHA@.')


def update_go_ebuild(ebuild: str, version: str, go_sum_uri_template: str) -> None:
    """
    Update a Go ebuild with the latest EGO_SUM content.

    Raises
    ------
    InvalidGoSumURITemplate
        If the URI template does not contain '@PV@' or '@SHA@'.
    """
    if '@PV@' not in go_sum_uri_template and '@SHA@' not in go_sum_uri_template:
        raise InvalidGoSumURITemplate
    sha = ''
    with contextlib.suppress(StopIteration):
        if (first_match :=
                next(y for y in (re.match(r'^SHA="([^"]+)"', x)
                                 for x in Path(ebuild).read_text(encoding='utf-8').splitlines())
                     if y is not None)):
            sha = first_match.group(1)
    uri = go_sum_uri_template.replace('@PV@', version).replace('@SHA@', sha)
    if not (r := get_content(uri)):
        return
    new_ego_sum_lines = (f'"{line}"'
                         for line in (re.sub(r' h1\:.*$', '', x) for x in r.text.splitlines()))
    with EbuildTempFile(ebuild) as temp_file, temp_file.open('w', encoding='utf-8') as tf:
        updated = False
        found_closing_bracket = False
        with Path(ebuild).open('r', encoding='utf-8') as f:
            for line in f:
                if line.startswith('EGO_SUM=(') and not updated:
                    logger.debug('Found EGO_SUM=( line.')
                    tf.write('EGO_SUM=(\n')
                    logger.debug('Writing new EGO_SUM content.')
                    for pkg in new_ego_sum_lines:
                        tf.write(f'\t{pkg}\n')
                    tf.write(')\n')
                    updated = True
                elif updated and not found_closing_bracket:
                    if line.strip() == ')':
                        logger.debug('Found closing bracket for EGO_SUM.')
                        found_closing_bracket = True
                else:
                    tf.write(line)



"""Checksum functions."""
from __future__ import annotations

from pathlib import Path
import re

from livecheck.utils import get_last_modified, hash_url
from livecheck.utils.portage import catpkg_catpkgsplit

from .utils import EbuildTempFile

__all__ = ('get_latest_checksum_package',)

PATTERN = re.compile(r'^DIST\s+(?P<file>\S+)\s+(?P<size>\d+)\s+BLAKE2B\s+'
                     r'(?P<blake2b>[a-fA-F0-9]+)\s+SHA512\s+(?P<sha512>[a-fA-F0-9]+)$')


def get_latest_checksum_package(url: str, ebuild: str, repo_root: str) -> tuple[str, str, str]:
    """Get the latest version of a package based on its checksum."""
    catpkg, _, _, version = catpkg_catpkgsplit(ebuild)
    manifest_file = Path(repo_root) / catpkg / 'Manifest'
    bn = Path(url).name

    with Path(manifest_file).open(encoding='utf-8') as f:
        for line in f:
            m = PATTERN.match(line)
            if m and m.group('file') == bn:
                blake2, sha512, _ = hash_url(url)
                if blake2 != m.group('blake2b') or sha512 != m.group('sha512'):
                    last_modified = get_last_modified(url)
                    return version, last_modified, url

    return '', '', ''


def update_checksum_metadata(ebuild: str, url: str, repo_root: str) -> None:
    """Update the checksum metadata in the Manifest file."""
    catpkg, _, _, _ = catpkg_catpkgsplit(ebuild)
    manifest_file = Path(repo_root) / catpkg / 'Manifest'
    blake2, sha512, size = hash_url(url)
    bn = Path(url).name

    with (
            EbuildTempFile(str(manifest_file)) as temp_file,
            temp_file.open('w', encoding='utf-8') as tf,
            Path(manifest_file).open('r', encoding='utf-8') as f,
    ):
        for line in f:
            m = PATTERN.match(line)
            if m and m.group('file') == bn:
                tf.write(f'DIST {bn} {size} BLAKE2B {blake2} SHA512 {sha512}\n')
            else:
                tf.write(line)



"""Portage utilities."""
from __future__ import annotations

from functools import cache
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING
import logging
import re

from portage.versions import catpkgsplit, vercmp
import portage

if TYPE_CHECKING:
    from collections.abc import Collection, Iterable, Mapping

    from livecheck.settings import LivecheckSettings

__all__ = ('P', 'catpkg_catpkgsplit', 'catpkgsplit2', 'compare_versions', 'fetch_ebuild',
           'get_distdir', 'get_first_src_uri', 'get_highest_matches', 'get_last_version',
           'get_repository_root_if_inside', 'remove_leading_zeros', 'sanitize_version',
           'unpack_ebuild')

P = portage.db[portage.root]['porttree'].dbapi
log = logging.getLogger(__name__)


def mask_version(cp: str, version: str, restrict_version: str | None = 'full') -> str:
    if restrict_version == 'major':
        return cp + ':' + re.sub(r'^(\d+).*', r'\1', version) + ':'
    if restrict_version == 'minor':
        return cp + ':' + re.sub(r'^(\d+\.\d+).*', r'\1', version) + ':'
    return cp


def get_highest_matches(names: Iterable[str], repo_root: Path | None,
                        settings: LivecheckSettings) -> list[str]:
    """Get the highest matching versions for an iterable of package names."""
    log.debug('Searching for %s.', ', '.join(names))
    result: dict[str, str] = {}
    for name in names:
        if not (matches := P.xmatch('match-all', name)):
            log.debug('Found no matches with xmatch("match-all").')
            continue
        for m in matches:
            # Check if the package structure is valid
            try:
                cp_a, _, _, version = catpkg_catpkgsplit(m)
            except ValueError:
                log.debug('Ignoring invalid package structure.')
                continue

            if repo_root and (actual_root := P.findname2(m)[1]) != str(repo_root):
                log.debug('Ignoring invalid repository root. Expected `%s` and received `%s`.',
                          repo_root, actual_root)
                continue

            if '9999' in version or not cp_a or not version:
                log.debug('Ignoring 9999 version.')
                continue

            restrict_version = settings.restrict_version.get(name, 'full')
            cp_mask = mask_version(cp_a, version, restrict_version)

            if cp_mask not in result or vercmp(version, result[cp_mask]):
                result[cp_mask] = version

    return [f'{cp}-{version}' for cp, version in result.items()]


CATPKGSPLIT_SIZE = 4
"""Size of the tuple returned by :py:func:`catpkgsplit`."""


@cache
def catpkgsplit2(atom: str) -> tuple[str | None, str, str, str]:
    """
    Split an atom string. This function always returns a four-string tuple.

    Parameters
    ----------
    atom : str
        String to split.

    Returns
    -------
    tuple[str | None, str, str, str]
        Tuple consisting of four strings. If category is not set, the first item is ``None``.

    Raises
    ------
    ValueError
        If :py:func:`catpkgsplit` returns ``None`` or a tuple of size not equal to 4.
    """
    result = catpkgsplit(atom)
    if result is None or len(result) != CATPKGSPLIT_SIZE:
        msg = f'Invalid atom: {atom}'
        raise ValueError(msg)

    return result[0], result[1], result[2], result[3]


@cache
def catpkg_catpkgsplit(atom: str) -> tuple[str, str, str, str]:
    """
    Split an atom string into category, package, and version, but also return CP.

    Returns
    -------
    tuple[str, str, str, str]
        Tuple consisting of CP, category, PN, and PV.
    """
    cat, pkg, ebuild_version, revision = catpkgsplit2(atom)
    assert cat is not None

    if revision and revision != 'r0':
        return f'{cat}/{pkg}', cat, pkg, f'{ebuild_version}-{revision}'

    return f'{cat}/{pkg}', cat, pkg, ebuild_version


def get_first_src_uri(match: str, search_dir: Path | None = None) -> str:
    """Get the first source URI for a match string (passed to :py:func:`P.aux_get`)."""
    try:
        if (found_uri := next((uri for uri in chain(
                *(x.split()
                  for x in map(str, P.aux_get(match, ['SRC_URI'], mytree=str(search_dir)))))
                               if uri.startswith(('http://', 'https://', 'mirror://', 'ftp://'))),
                              None)):
            return found_uri
    except KeyError:
        pass
    return ''


def get_repository_root_if_inside(directory: Path) -> tuple[str, str]:
    """Get the repository root if the current working directory is inside a repository."""
    # Get Portage configuration
    settings = portage.config(clone=portage.settings)

    # Get repositories from the settings object
    repos = [*settings['PORTDIR_OVERLAY'].split(), settings['PORTDIR']]

    # Normalize the directory path to check
    directory = directory.resolve(strict=True)
    directory_str = str(directory) + '/'
    selected_repo_root = ''
    selected_repo_name = ''

    # Check each repository
    for repo_root in repos:
        if Path(repo_root).is_dir():
            repo_root_ = str(Path(repo_root).resolve(strict=True))
            # Check if the directory is inside the repository root
            # Select the most specific repository (deepest path)
            if directory_str.startswith(f'{repo_root_}/') and not selected_repo_root:
                selected_repo_root = repo_root_
                selected_repo_name = Path(repo_root_).name

    if '/local/' in directory_str and '/local/' not in str(selected_repo_root):
        return '', ''

    # Return the most specific repository root, if found
    return selected_repo_root, selected_repo_name


def is_version_development(version: str) -> bool:
    return bool(re.search(r'(alpha|beta|pre|dev|rc)', version, re.IGNORECASE))


def remove_initial_match(a: str, b: str) -> str:
    i = 0
    while i < len(a) and i < len(b) and a[i] == b[i]:
        i += 1
    return a[i:]


def extract_version(s: str, repo: str) -> str:
    # force convert to string to avoid a int object has no attribute lower
    s = str(s).lower().strip()
    # check if first word of s is equal to repo and remove repo from s

    s = remove_initial_match(s, repo.lower())
    s.strip()

    if m := re.search(r'[-_]?([0-9][0-9\._-].*)', s):
        return m.group(1).strip()

    m = re.search(r'(?:^|[^-_])(\d.*)', s)
    return m.group(1).strip() if m else ''


def sanitize_version(ver: str, repo: str = '') -> str:
    """Sanitise a version string."""
    ver = extract_version(ver, repo)
    ver = normalize_version(ver)
    return remove_leading_zeros(ver)


def remove_leading_zeros(ver: str) -> str:
    """Check if version has a date string like ``2022.12.26``."""
    if not re.match(r'\d{4}|\d{2}\.\d{2}\.\d{2}', ver):
        return ver
    if match := re.match(r'(\d+)\.(\d+)\.(\d+)(.*)', ver):
        a, b, c, suffix = match.groups()
        return f'{int(a)}.{int(b)}.{int(c)}{suffix}'
    return ver


def normalize_version(ver: str) -> str:  # noqa: PLR0911
    """
    Normalize version string to Gentoo Ebuild format.

    See Also
    --------
    `Guide <https://devmanual.gentoo.org/ebuild-writing/file-format/ebuild-format.html>`_

    Parameters
    ----------
    ver : str
        The version string to normalize.

    Returns
    -------
    str
        The normalized version string.
    """
    i = 0
    sep = '._-'
    while i < len(ver) and (ver[i].isdigit() or ver[i] in sep):
        if ver[i] in sep:
            sep = ver[i]
        i += 1
    main = re.sub(r'[-_]', '.', ver[:i])
    suf = ver[i:]

    if not (main := main.rstrip('.')):
        return ver

    suf = re.sub(r'[-_\. ]', '', suf)
    if suf.isdigit():
        return f'{main}.{suf}'

    if m := re.match(r'^([A-Za-z]+)([0-9]+)?', suf):
        letters, digits = m.groups()
    else:
        letters, digits = '', ''

    if digits:
        if letters == 'a':
            letters = 'alpha'
        if letters == 'b':
            letters = 'beta'
    if digits == '0':
        digits = ''

    if letters in {'test', 'dev'}:
        letters = 'beta'
    if letters.startswith(('pl', 'patchlevel')):
        letters = 'p'

    allowed = ('pre', 'beta', 'rc', 'p', 'alpha', 'post')

    if letters in allowed:
        if letters in ('post'):
            letters = 'p'
        if digits:
            return f'{main}_{letters}{digits}'
        return f'{main}_{letters}'

    # Single-letter suffix with no digits -> preserve as lowercase
    if len(letters) == 1 and not digits:
        return f'{main}{letters}'
    # No recognized suffix
    if digits:
        # Just attach the digits directly (e.g. "1.2.3" + "4")
        return f'{main}{digits}'
    # If the version ends with a letter like 1.2.20a (and not recognized),
    # the requirement says "it is preserved" only if it is exactly a single letter.
    # For multi-letter unknown suffix -> discard.
    return main


def compare_versions(old: str, new: str) -> bool:
    """
    Compare two version strings.

    Returns
    -------
    bool
        ``True`` if the old version is less than the new version, ``False`` otherwise.
    """
    return bool(vercmp(old, new) == -1)


def get_distdir() -> Path:
    """
    Get the ``DISTDIR`` path from Portage settings.

    Falls back to default ``/var/cache/distfiles``.
    """
    settings = portage.config(clone=portage.settings)
    if distdir := settings.get('DISTDIR'):
        return Path(distdir)
    return Path('/var/cache/distfiles')


def fetch_ebuild(ebuild_path: str) -> bool:
    """Perform ``ebuild fetch`` operation."""
    settings = portage.config(clone=portage.settings)
    return bool(portage.doebuild(ebuild_path, 'fetch', settings=settings, tree='porttree') == 0)


def digest_ebuild(ebuild_path: str) -> bool:
    """Perform ``ebuild digest`` operation."""
    settings = portage.config(clone=portage.settings)
    return bool(portage.doebuild(ebuild_path, 'digest', settings=settings, tree='porttree') == 0)


def unpack_ebuild(ebuild_path: str) -> str:
    """Perform ``ebuild unpack`` operation and return the WORKDIR path."""
    settings = portage.config(clone=portage.settings)

    if portage.doebuild(ebuild_path, 'clean', settings=settings, tree='porttree') != 0:
        return ''

    if portage.doebuild(ebuild_path, 'unpack', settings=settings, tree='porttree') != 0:
        return ''

    workdir = Path(settings['WORKDIR'])

    if workdir.exists() and workdir.is_dir():
        return str(workdir)

    return ''


def get_last_version(results: Collection[Mapping[str, str]], repo: str, ebuild: str,
                     settings: LivecheckSettings) -> dict[str, str]:
    """
    Get the latest version from the results.

    Parameters
    ----------
    results : Collection[Mapping[str, str]]
    repo : str
    ebuild : str
    settings : LivecheckSettings
    """
    log.debug('Result count: %d', len(results))

    catpkg, _, _, ebuild_version = catpkg_catpkgsplit(ebuild)
    last_version: dict[str, str] = {}

    for result in results:
        tag = version = result['tag']
        if tf := settings.transformations.get(catpkg, None):
            version = tf(tag)
            log.debug('Applying transformation %s -> %s', tag, version)
        if catpkg in settings.regex_version:
            regex, replace = settings.regex_version[catpkg]
            version = re.sub(regex, replace, version)
            log.debug('Applying regex %s -> %s', tag, version)
        else:
            version = sanitize_version(version, repo)
            log.debug('Convert Tag: %s -> %s', tag, version)
        if not version:
            continue
        # skip version extraneous without dots, example Post120ToMaster
        if ebuild_version.count('.') > 1 and version.count('.') == 0:
            log.debug('Skip version without dots: %s', version)
            continue
        # Check valid version
        try:
            _, _, _, _ = catpkg_catpkgsplit(f'{catpkg}-{version}')
        except ValueError:
            log.debug('Skip non-version tag: %s', version)
            continue
        if not version.startswith(settings.restrict_version_process):
            continue
        if accept_version(ebuild_version, version, catpkg, settings):
            last = last_version.get('version', '')
            if not last or compare_versions(last, version):
                last_version = dict(result).copy()
                last_version['version'] = version

    if not last_version:
        log.debug('No new update for %s.', ebuild)

    return last_version


def accept_version(ebuild_version: str, version: str, catpkg: str,
                   settings: LivecheckSettings) -> bool:
    """
    Determine if version is acceptable for the given CP.

    Parameters
    ----------
    ebuild_version : str
        The version of the ebuild.
    version : str
        The version to check against the ebuild version.
    catpkg : str
        The category/package name in the format 'category/package'.
    settings : LivecheckSettings
    """
    stable_version = settings.stable_version.get(catpkg, '')
    if is_version_development(ebuild_version) or settings.is_devel(catpkg) or (
            stable_version and re.match(stable_version, version)):
        return True

    return not (is_version_development(version) or
                (stable_version and not re.match(stable_version, version)))



"""Utility functions."""
from __future__ import annotations

from .assertions import assert_not_none
from .misc import check_program
from .requests import TextDataResponse, get_content, get_last_modified, hash_url, session_init
from .string import dash_to_underscore, dotize, extract_sha, is_sha, prefix_v

__all__ = ('TextDataResponse', 'assert_not_none', 'check_program', 'dash_to_underscore', 'dotize',
           'extract_sha', 'get_content', 'get_last_modified', 'hash_url', 'is_sha', 'prefix_v',
           'session_init')



"""Composer functions."""
from __future__ import annotations

from typing import TYPE_CHECKING
import logging
import subprocess as sp

from livecheck.utils import check_program

from .utils import build_compress, remove_url_ebuild, search_ebuild

if TYPE_CHECKING:
    from collections.abc import Mapping

__all__ = ('check_composer_requirements', 'remove_composer_url', 'update_composer_ebuild')

log = logging.getLogger(__name__)


def remove_composer_url(ebuild_content: str) -> str:
    """Remove the URL for the vendor tarball from the ebuild content."""
    return remove_url_ebuild(ebuild_content, '-vendor.tar.xz')


def update_composer_ebuild(ebuild: str, path: str | None,
                           fetchlist: Mapping[str, tuple[str, ...]]) -> None:
    """Update a Composer package ebuild."""
    composer_path, temp_dir = search_ebuild(ebuild, 'composer.json', path)
    if not composer_path:
        return

    try:
        sp.run(('composer', '--no-interaction', '--no-scripts', 'install'),
               cwd=composer_path,
               check=True)
    except sp.CalledProcessError:
        log.exception("Error running 'composer'.")
        return

    build_compress(temp_dir, composer_path, 'vendor', '-vendor.tar.xz', fetchlist)


def check_composer_requirements() -> bool:
    """Check if Composer is installed."""
    if not check_program('composer', ['--version']):
        log.error('composer is not installed')
        return False
    return True



"""General utilities for special handling."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING
import logging
import os
import tarfile
import tempfile

from livecheck.utils.portage import get_distdir, unpack_ebuild
from platformdirs import user_cache_dir

if TYPE_CHECKING:
    from collections.abc import Collection, Mapping

__all__ = ('EbuildTempFile', 'build_compress', 'get_archive_extension', 'get_project_path',
           'remove_url_ebuild', 'search_ebuild')

logger = logging.getLogger(__name__)


def get_project_path(package_name: str) -> Path:
    """Get the project path for a given package name."""
    return Path(user_cache_dir('livecheck')) / package_name


def remove_url_ebuild(ebuild: str, remove: str) -> str:
    """Remove URLs from the ebuild content."""
    lines = ebuild.split('\n')
    filtered_lines = []
    for line in lines:
        original_line = line
        stripped_line = line.strip()
        if not stripped_line or stripped_line.startswith('#'):
            filtered_lines.append(original_line)
            continue
        if remove in stripped_line and stripped_line.strip(' "\'').endswith(remove):
            if (stripped_line.endswith(('"', "'"))
                    and (stripped_line.count('"') == 1 or stripped_line.count("'") == 1)):
                filtered_lines.append(stripped_line[-1])
            continue
        filtered_lines.append(original_line)
    return '\n'.join(filtered_lines)


def search_ebuild(ebuild: str, archive: str, path: str | None = None) -> tuple[str, str]:
    """Search for an archive in the unpacked ebuild directory."""
    temp_dir = unpack_ebuild(ebuild)
    if not temp_dir:
        logger.warning('Error unpacking the ebuild.')
        return '', ''

    if path:
        # Search first directory in temp_dir
        for root, _, _ in os.walk(temp_dir):
            # check if relative path is in the root
            if path in root:
                return root, temp_dir
    else:
        for root, _, files in os.walk(temp_dir):
            if archive in files:
                return root, temp_dir

    logger.error('Error searching the `%s` inside package.', archive)

    return '', ''


def build_compress(temp_dir: str, base_dir: str, directory: str, extension: str,
                   fetchlist: Mapping[str, Collection[str]]) -> bool:
    """Build dist archive."""
    vendor_dir = Path(base_dir) / directory
    if not vendor_dir.exists():
        logger.warning('The directory vendor was not created.')
        return False

    if not (filename := next(iter(fetchlist.keys()), None)):
        return False

    if not (archive_ext := get_archive_extension(filename)):
        logger.warning('Invalid extension.')
        return False

    if extension in filename:
        vendor_archive_name = filename
    else:
        base_name = filename[:-len(archive_ext)]
        vendor_archive_name = f'{base_name}{extension}'
    vendor_archive_path = get_distdir() / vendor_archive_name

    vendor_path = Path(base_dir).resolve()
    base_path = Path(temp_dir).resolve()

    relative_path = vendor_path.relative_to(base_path) / directory

    with tarfile.open(vendor_archive_path, 'w:xz') as tar:
        tar.add(vendor_dir, arcname=str(relative_path))

    return True


def get_archive_extension(filename: str) -> str:
    """Get archive extension from a filename."""
    filename = filename.lower()
    for ext in ('tar.gz', 'tar.xz', 'tar.bz2', 'tar.lz', 'tar.zst', 'tc.gz', 'tar.z', 'gz', 'xz',
                'zip', 'tbz2', 'bz2', 'tbz', 'txz', 'tar', 'tgz', 'rar', '7z'):
        if filename.endswith(f'.{ext}'):
            return '.' + ext

    return ''


class EbuildTempFile:
    """Ebuild temporary file context manager."""
    def __init__(self, ebuild: str) -> None:
        self.ebuild = Path(ebuild)
        self.temp_file: Path | None = None

    def __enter__(self) -> Path:
        """Create a temporary file."""
        self.temp_file = Path(
            tempfile.NamedTemporaryFile(mode='w',
                                        prefix=self.ebuild.stem,
                                        suffix=self.ebuild.suffix,
                                        delete=False,
                                        dir=self.ebuild.parent,
                                        encoding='utf-8').name)
        return self.temp_file

    def __exit__(self, exc_type: object, exc_value: BaseException | None,
                 traceback: object) -> None:
        """Handle the context exit."""
        if exc_type is None:
            if not self.temp_file or not self.temp_file.exists() or self.temp_file.stat(
            ).st_size == 0:
                logger.error('The temporary file is empty or missing.')
                return
            self.ebuild.unlink(missing_ok=True)
            self.temp_file.rename(self.ebuild)
            self.ebuild.chmod(0o0644)
        if self.temp_file and self.temp_file.exists():
            self.temp_file.unlink(missing_ok=True)


def log_unhandled_commit(catpkg: str, src_uri: str) -> None:
    logger.warning('Unhandled commit: %s SRC_URI: %s', catpkg, src_uri)



from __future__ import annotations

from typing import TYPE_CHECKING

from livecheck.special.handlers import (
    handle_bsnes_hd,
    handle_cython_post_suffix,
    handle_glabels,
    handle_outfox,
    handle_outfox_serenity,
    handle_pl,
)

if TYPE_CHECKING:
    from pytest_mock import MockerFixture


def test_handle_glabels_returns_modified_version_when_hash_date_present(
        mocker: MockerFixture) -> None:
    mock_get_latest = mocker.patch('livecheck.special.handlers.get_latest_github_commit2',
                                   return_value=('abcdef', '20240601'))
    handle_glabels.cache_clear()
    result = handle_glabels('some-ver')
    assert result == '3.99_p20240601'
    mock_get_latest.assert_called_once_with('jimevins', 'glabels-qt', 'master')


def test_handle_glabels_returns_input_when_hash_date_missing(mocker: MockerFixture) -> None:
    mock_get_latest = mocker.patch('livecheck.special.handlers.get_latest_github_commit2',
                                   return_value=('abcdef', None))
    handle_glabels.cache_clear()
    result = handle_glabels('orig-ver')
    assert result == 'orig-ver'
    mock_get_latest.assert_called_once_with('jimevins', 'glabels-qt', 'master')


def test_handle_cython_post_suffix_replaces_post_with_dot() -> None:
    input_version = '0.29.21.post1'
    result = handle_cython_post_suffix(input_version)
    assert result == '0.29.21.1'


def test_handle_cython_post_suffix_no_post() -> None:
    input_version = '0.29.21'
    result = handle_cython_post_suffix(input_version)
    assert result == '0.29.21'


def test_handle_cython_post_suffix_multiple_post() -> None:
    input_version = '1.0.post2.post3'
    result = handle_cython_post_suffix(input_version)
    assert result == '1.0.2.3'


def test_handle_outfox_with_pre_suffix() -> None:
    input_version = '0.4.18-pre0001'
    result = handle_outfox(input_version)
    assert result == '0.4.18_p1'


def test_handle_outfox_with_pre_suffix_multiple_zeros() -> None:
    input_version = '0.4.18-pre0000123'
    result = handle_outfox(input_version)
    assert result == '0.4.18_p123'


def test_handle_outfox_without_pre_suffix() -> None:
    input_version = '0.4.18'
    result = handle_outfox(input_version)
    assert result == '0.4.18'


def test_handle_outfox_with_pre_suffix_and_trailing_text() -> None:
    input_version = '0.4.18-pre0001-extra'
    result = handle_outfox(input_version)
    assert result == '0.4.18_p1-extra'


def test_handle_outfox_serenity_replaces_s_with_dot() -> None:
    input_version = '1s2s3'
    result = handle_outfox_serenity(input_version)
    assert result == '1.2.3'


def test_handle_outfox_serenity_no_s_in_input() -> None:
    input_version = '1.2.3'
    result = handle_outfox_serenity(input_version)
    assert result == '1.2.3'


def test_handle_outfox_serenity_multiple_consecutive_s() -> None:
    input_version = '1ss2'
    result = handle_outfox_serenity(input_version)
    assert result == '1..2'


def test_handle_outfox_serenity_s_at_start_and_end() -> None:
    input_version = 's1.2.3s'
    result = handle_outfox_serenity(input_version)
    assert result == '.1.2.3.'


def test_handle_outfox_serenity_empty_string() -> None:
    input_version = ''
    result = handle_outfox_serenity(input_version)
    assert not result


def test_handle_bsnes_hd_basic() -> None:
    # beta_115_10h2 -> 115.10_beta
    input_version = 'beta_115_10h2'
    result = handle_bsnes_hd(input_version)
    assert result == '115.10_beta'


def test_handle_bsnes_hd_no_h() -> None:
    # beta_2_3 -> 2.3_beta
    input_version = 'beta_2_3'
    result = handle_bsnes_hd(input_version)
    assert result == '2.3_beta'


def test_handle_bsnes_hd_multiple_h() -> None:
    # beta_5_12h3h4 -> 5.12_beta
    input_version = 'beta_5_12h3h4'
    result = handle_bsnes_hd(input_version)
    assert result == '5.12_beta'


def test_handle_bsnes_hd_minor_with_h() -> None:
    # beta_7_8h1 -> 7.8_beta
    input_version = 'beta_7_8h1'
    result = handle_bsnes_hd(input_version)
    assert result == '7.8_beta'


def test_handle_pl_basic() -> None:
    input_version = '1.2.3-pl4'
    result = handle_pl(input_version)
    assert result == '1.2.3.4'


def test_handle_pl_with_v_prefix() -> None:
    input_version = 'v2.5.7-pl10'
    result = handle_pl(input_version)
    assert result == '2.5.7.10'


def test_handle_pl_no_pl_suffix() -> None:
    input_version = '1.2.3'
    result = handle_pl(input_version)
    assert not result


def test_handle_pl_incorrect_format() -> None:
    input_version = '1.2-pl3'
    result = handle_pl(input_version)
    assert not result


def test_handle_pl_non_numeric_pl() -> None:
    input_version = '1.2.3-plx'
    result = handle_pl(input_version)
    assert not result



"""SourceHut utility functions."""
from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING
from urllib.parse import urlparse
import re

from defusedxml import ElementTree as ET  # noqa: N817
from livecheck.utils import get_content, is_sha
from livecheck.utils.portage import catpkg_catpkgsplit, get_last_version

if TYPE_CHECKING:
    from livecheck.settings import LivecheckSettings

__all__ = ('SOURCEHUT_METADATA', 'get_latest_sourcehut', 'get_latest_sourcehut_commit',
           'get_latest_sourcehut_metadata', 'get_latest_sourcehut_package', 'is_sourcehut')

SOURCEHUT_DOWNLOAD_URL = 'https://%s/%s/%s/refs/rss.xml'
SOURCEHUT_COMMIT_URL = 'https://%s/%s/%s/log/%s/rss.xml'
SOURCEHUT_METADATA = 'sourcehut'


def extract_owner_repo(url: str) -> tuple[str, str, str]:
    if m := re.match(
            r'^(https?://)?(?P<d>(?:(?:git|hg)\.)?sr\.ht)/(?P<u>~?[^/]+)/(?P<p>[^/]+)(/.*)?', url):
        return m.group('d'), m.group('u'), m.group('p')
    return '', '', ''


def get_latest_sourcehut_package(url: str, ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a SourceHut package."""
    domain, owner, repo = extract_owner_repo(url)
    if not owner or not repo:
        return ''

    url = SOURCEHUT_DOWNLOAD_URL % (domain, owner, repo)

    if not (r := get_content(url)):
        return ''

    results: list[dict[str, str]] = []
    for item in ET.fromstring(r.text).findall('channel/item'):
        guid = item.find('guid')
        if version := guid.text.split('/')[-1] if guid is not None and guid.text else '':
            results.append({'tag': version})

    if last_version := get_last_version(results, repo, ebuild, settings):
        return last_version['version']
    return ''


def get_latest_sourcehut_commit(url: str, branch: str = 'master') -> tuple[str, str]:
    """Get the latest commit hash and date from a SourceHut repository."""
    domain, owner, repo = extract_owner_repo(url)
    if not owner or not repo:
        return '', ''

    url = SOURCEHUT_COMMIT_URL % (domain, owner, repo, branch)

    if not (r := get_content(url)):
        return '', ''

    guid = ET.fromstring(r.text).find('channel/item/guid')
    pubdate = ET.fromstring(r.text).find('channel/item/pubDate')
    commit = guid.text.split('/')[-1] if guid is not None and guid.text else ''
    date = pubdate.text if pubdate is not None and pubdate.text else ''

    try:
        dt = datetime.strptime(date, '%a, %d %b %Y %H:%M:%S %z')
        formatted_date = dt.strftime('%Y%m%d')
    except ValueError:
        formatted_date = ''
    return commit, formatted_date


def is_sourcehut(url: str) -> bool:
    """Check if the given URL is a SourceHut repository."""
    return bool(extract_owner_repo(url)[0])


def get_branch(url: str, ebuild: str, settings: LivecheckSettings) -> str:
    catpkg, _, _, _ = catpkg_catpkgsplit(ebuild)

    # get branch from url
    parts = url.strip('/').split('/')
    if len(parts) >= 3 and parts[-3] == 'log':  # noqa: PLR2004
        return parts[-2]

    # get branch from settings
    if branch := settings.branches.get(catpkg, ''):
        return branch

    # default branch is master
    if is_sha(urlparse(url).path):
        return 'master'

    return ''


def get_latest_sourcehut(url: str, ebuild: str, settings: LivecheckSettings, *,
                         force_sha: bool) -> tuple[str, str, str]:
    """Get the latest version and commit hash from a SourceHut repository."""
    last_version = top_hash = hash_date = ''

    if (branch := get_branch(url, ebuild, settings)):
        top_hash, hash_date = get_latest_sourcehut_commit(url, branch)
        if not force_sha:
            top_hash = ''
    else:
        last_version = get_latest_sourcehut_package(url, ebuild, settings)

    return last_version, top_hash, hash_date


def get_latest_sourcehut_metadata(remote: str, ebuild: str, settings: LivecheckSettings) -> str:
    """Get the latest version of a SourceHut package from its metadata."""
    if not (last_version := get_latest_sourcehut_package(f'https://git.sr.ht/{remote}', ebuild,
                                                         settings)):
        last_version = get_latest_sourcehut_package(f'https://hg.sr.ht/{remote}', ebuild, settings)

    return last_version



"""Github Gist functions."""
from __future__ import annotations

from datetime import datetime
import re

from livecheck.utils import get_content

__all__ = ('get_latest_gist_package', 'is_gist')

GIST_COMMIT_URL = 'https://api.github.com/gists/%s'


def extract_id(url: str) -> str:
    m = re.search(r'https?://gist\.github\.com/(?:[^/]+/)?([a-f0-9]+)', url)
    if not m:
        m = re.search(r'https?://gist\.githubusercontent\.com/[^/]+/([a-f0-9]+)/', url)
    return m.group(1) if m else ''


def get_latest_gist_package(url: str) -> tuple[str, str]:
    """Get the latest version of a Gist."""
    if not (gist_id := extract_id(url)):
        return '', ''

    url = GIST_COMMIT_URL % (gist_id)
    if not (r := get_content(url)):
        return '', ''

    history = r.json().get('history', [])
    if not history:
        return '', ''
    latest = max(history, key=lambda x: x.get('committed_at', ''))

    d = latest.get('committed_at')
    try:
        dt = datetime.fromisoformat(d.replace('Z', '+00:00'))
        formatted_date = dt.strftime('%Y%m%d')
    except ValueError:
        formatted_date = d[:10]

    return latest.get('version'), formatted_date


def is_gist(url: str) -> bool:
    """Check if the URL is to a Gist."""
    return bool(extract_id(url))



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from livecheck.special.gist import extract_id, get_latest_gist_package, is_gist
import pytest

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

test_cases = {
    'test_extract_id_github_com': {
        'url': 'https://gist.github.com/username/abcdef1234567890',
        'expected': 'abcdef1234567890',
        'gist': True
    },
    'test_extract_id_githubusercontent_com': {
        'url': 'https://gist.githubusercontent.com/username/abcdef1234567890/raw/filename',
        'expected': 'abcdef1234567890',
        'gist': True
    },
    'test_extract_id_invalid_url': {
        'url': 'https://example.com/username/abcdef1234567890',
        'expected': '',
        'gist': False
    },
    'test_extract_id_no_id': {
        'url': 'https://gist.github.com/username/',
        'expected': '',
        'gist': False
    },
    'test_extract_id_empty_string': {
        'url': '',
        'expected': '',
        'gist': False
    },
    'test_ntasos': {
        'url': 'https://gist.github.com/ntasos/8d1c7e2d2f8f5f1f3c9f',
        'expected': '8d1c7e2d2f8f5f1f3c9f',
        'gist': True
    },
    'test_bitrock': {
        'url':
            'https://gist.githubusercontent.com/mickael9/0b902da7c13207d1b86e/raw/ef65b6583c5cf077fb897a9a028781f577fed9ac/bitrock-unpacker.tcl',
        'expected':
            '0b902da7c13207d1b86e',
        'gist':
            True
    },
}


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_extract_id(test_case: dict[str, str]) -> None:
    assert extract_id(test_case['url']) == test_case['expected']


@pytest.mark.parametrize('test_case', test_cases.values(), ids=test_cases.keys())
def test_is_gist(test_case: dict[str, Any]) -> None:
    assert is_gist(test_case['url']) == test_case['gist']


def test_get_latest_gist_package_valid(mocker: MockerFixture) -> None:
    url = 'https://gist.github.com/username/abcdef1234567890'
    mock_response = mocker.Mock()
    mock_response.json.return_value = {
        'history': [{
            'version': 'v1',
            'committed_at': '2023-12-31T23:59:59Z'
        }, {
            'version': 'v2',
            'committed_at': '2024-01-01T12:00:00Z'
        }]
    }
    mocker.patch('livecheck.special.gist.get_content', return_value=mock_response)
    version, date = get_latest_gist_package(url)
    assert version == 'v2'
    assert date == '20240101'


def test_get_latest_gist_package_no_gist_id(mocker: MockerFixture) -> None:
    url = 'https://example.com/not-a-gist'
    version, date = get_latest_gist_package(url)
    assert not version
    assert not date


def test_get_latest_gist_package_no_content(mocker: MockerFixture) -> None:
    url = 'https://gist.github.com/username/abcdef1234567890'
    mocker.patch('livecheck.special.gist.get_content', return_value=None)
    version, date = get_latest_gist_package(url)
    assert not version
    assert not date


def test_get_latest_gist_package_no_history(mocker: MockerFixture) -> None:
    url = 'https://gist.github.com/username/abcdef1234567890'
    mock_response = mocker.Mock()
    mock_response.json.return_value = {'history': []}
    mocker.patch('livecheck.special.gist.get_content', return_value=mock_response)
    version, date = get_latest_gist_package(url)
    assert not version
    assert not date


def test_get_latest_gist_package_invalid_date(mocker: MockerFixture) -> None:
    url = 'https://gist.github.com/username/abcdef1234567890'
    mock_response = mocker.Mock()
    mock_response.json.return_value = {'history': [{'version': 'v1', 'committed_at': 'not-a-date'}]}
    mocker.patch('livecheck.special.gist.get_content', return_value=mock_response)
    version, date = get_latest_gist_package(url)
    assert version == 'v1'
    assert date == 'not-a-date'
